from __future__ import annotations

from coredis.commands import ResponseCallback
from coredis.response.types import (
    StreamEntry,
    StreamInfo,
    StreamPending,
    StreamPendingExt,
)
from coredis.response.utils import flat_pairs_to_dict, flat_pairs_to_ordered_dict
from coredis.typing import Any, AnyStr, Mapping, Optional, Tuple, Union
from coredis.utils import EncodingInsensitiveDict


class StreamRangeCallback(ResponseCallback):
    def transform(self, response: Any, **options: Any) -> Tuple[StreamEntry, ...]:
        return tuple(
            StreamEntry(r[0], flat_pairs_to_ordered_dict(r[1])) for r in response
        )


class ClaimCallback(ResponseCallback):
    def transform(
        self, response: Any, **options: Any
    ) -> Union[Tuple[AnyStr, ...], Tuple[StreamEntry, ...]]:
        if options.get("justid") is not None:
            return tuple(response)
        else:
            return StreamRangeCallback()(response)


class AutoClaimCallback(ResponseCallback):
    def transform(
        self, response: Any, **options: Any
    ) -> Union[
        Tuple[AnyStr, Tuple[AnyStr, ...]],
        Tuple[AnyStr, Tuple[StreamEntry, ...], Tuple[AnyStr, ...]],
    ]:
        if options.get("justid") is not None:
            return response[0], tuple(response[1])
        else:
            return (
                response[0],
                StreamRangeCallback()(response[1]),
                tuple(response[2]) if len(response) > 2 else (),
            )


class MultiStreamRangeCallback(ResponseCallback):
    def transform_3(
        self, response: Any, **options: Any
    ) -> Optional[Mapping[AnyStr, Tuple[StreamEntry, ...]]]:
        if response:
            mapping = {}

            for stream_id, entries in response.items():
                mapping[stream_id] = tuple(
                    StreamEntry(r[0], flat_pairs_to_ordered_dict(r[1])) for r in entries
                )

            return mapping
        return None

    def transform(
        self, response: Any, **options: Any
    ) -> Optional[Mapping[AnyStr, Tuple[StreamEntry, ...]]]:
        if response:
            mapping = {}

            for stream_id, entries in response:
                mapping[stream_id] = tuple(
                    StreamEntry(r[0], flat_pairs_to_ordered_dict(r[1])) for r in entries
                )

            return mapping
        return None


class PendingCallback(ResponseCallback):
    def transform(
        self, response: Any, **options: Any
    ) -> Union[StreamPending, Tuple[StreamPendingExt, ...]]:
        if not options.get("count"):
            return StreamPending(
                response[0],
                response[1],
                response[2],
                flat_pairs_to_ordered_dict(response[3:]),
            )
        else:
            return tuple(
                StreamPendingExt(sub[0], sub[1], sub[2], sub[3]) for sub in response
            )


class XInfoCallback(ResponseCallback):
    def transform(
        self, response: Any, **options: Any
    ) -> Tuple[Mapping[AnyStr, AnyStr], ...]:
        return tuple(flat_pairs_to_dict(row) for row in response)


class StreamInfoCallback(ResponseCallback):
    def transform(self, response: Any, **options: Any) -> StreamInfo:
        res = EncodingInsensitiveDict(flat_pairs_to_dict(response))
        if not options.get("full"):

            k1 = "first-entry"
            kn = "last-entry"
            e1, en = None, None
            if len(res.get(k1, [])) > 0:
                v = res.get(k1)
                e1 = StreamEntry(v[0], flat_pairs_to_ordered_dict(v[1]))
                res.pop(k1)

            if len(res.get(kn, [])) > 0:
                v = res.get(kn)
                en = StreamEntry(v[0], flat_pairs_to_ordered_dict(v[1]))
                res.pop(kn)
            res.update({"first-entry": e1, "last-entry": en})
        else:
            groups = res.get("groups")
            if groups:
                res.update({"groups": flat_pairs_to_dict(groups)})
            res.update(
                {
                    "entries": tuple(
                        StreamEntry(k[0], flat_pairs_to_ordered_dict(k[1]))
                        for k in res.get("entries", [])
                    )
                }
            )
        stream_info: StreamInfo = {
            "first-entry": res.get("first-entry"),
            "last-entry": res.get("last-entry"),
            "length": res["length"],
            "radix-tree-keys": res["radix-tree-keys"],
            "radix-tree-nodes": res["radix-tree-nodes"],
            "groups": res["groups"],
            "last-generated-id": res["last-generated-id"],
            "max-deleted-entry-id": str(res.get("max-deleted-entry-id")),
            "entries-added": int(res.get("entries-added", 0)),
            "recorded-first-entry-id": str(res.get("recorded-first-entry-id")),
            "entries-read": int(res.get("entries-read", 0)),
            "entries": res.get("entries"),
        }
        return stream_info



from __future__ import annotations

from coredis.commands import ResponseCallback
from coredis.typing import Any, AnyStr, Set, Tuple, Union


class SScanCallback(ResponseCallback):
    def transform(self, response: Any, **options: Any) -> Tuple[int, Set[AnyStr]]:
        cursor, r = response

        return int(cursor), set(r)


class ItemOrSetCallback(ResponseCallback):
    def transform(self, response: Any, **options: Any) -> Union[AnyStr, Set[AnyStr]]:
        if options.get("count"):
            if isinstance(response, set):
                return response
            return response and set(response)
        else:
            return response



"""
coredis.response.callbacks
--------------------------
"""

from __future__ import annotations

import datetime
import itertools
from abc import ABC, abstractmethod

from coredis.exceptions import ClusterResponseError
from coredis.typing import (
    Any,
    Callable,
    Generic,
    List,
    Mapping,
    Optional,
    ParamSpec,
    ResponseType,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
)

R = TypeVar("R")
P = ParamSpec("P")

_RESP = TypeVar("_RESP", bound=ResponseType)
_RESP3 = TypeVar("_RESP3", bound=ResponseType)


class ResponseCallback(ABC, Generic[_RESP, _RESP3, R]):
    def __call__(self, response: Any, version: int = 2, **kwargs: Any) -> R:
        if version == 3:
            return self.transform_3(response, **kwargs)
        return self.transform(response, **kwargs)

    @abstractmethod
    def transform(self, response: Union[_RESP, _RESP3], **kwargs: Any) -> R:
        pass

    def transform_3(self, response: _RESP3, **kwargs: Any) -> R:
        return self.transform(response, **kwargs)


class ClusterMultiNodeCallback(ABC):
    def __call__(
        self, responses: Mapping[str, Any], version: int = 2, **kwargs: Any
    ) -> Any:
        if version == 3:
            return self.combine_3(responses, **kwargs)
        return self.combine(responses, **kwargs)

    @abstractmethod
    def combine(self, responses: Mapping[str, Any], **kwargs: Any) -> Any:
        pass

    def combine_3(self, responses: Mapping[str, Any], **kwargs: Any) -> Any:
        return self.combine(responses, **kwargs)


class ClusterBoolCombine(ClusterMultiNodeCallback):
    def combine(self, responses: Mapping[str, Any], **kwargs: Any) -> bool:
        values = tuple(responses.values())
        assert (isinstance(value, bool) for value in values)
        return all(values)


class ClusterAlignedBoolsCombine(ClusterMultiNodeCallback):
    def combine(self, responses: Mapping[str, Any], **kwargs: Any) -> Tuple[bool, ...]:
        return tuple(all(k) for k in zip(*responses.values()))


class ClusterEnsureConsistent(ClusterMultiNodeCallback):
    def __init__(self, ensure_consistent=True):
        self.ensure_consistent = ensure_consistent

    def combine(self, responses: Mapping[str, Any], **kwargs: Any) -> Any:
        values = tuple(responses.values())
        if self.ensure_consistent and len(set(values)) != 1:
            raise ClusterResponseError(
                "Inconsistent response from cluster nodes", responses
            )
        else:
            return values and values[0]


class ClusterMergeSets(ClusterMultiNodeCallback):
    def combine(self, responses: Mapping[str, Any], **kwargs: Any) -> Set:
        return set(itertools.chain(*responses.values()))


class SimpleStringCallback(ResponseCallback):
    def __init__(self, raise_on_error: Optional[Type[Exception]] = None):
        self.raise_on_error = raise_on_error

    def transform(self, response: Any, **options: Any) -> Any:
        success = response and (response in {"OK", b"OK"})
        if not success and self.raise_on_error:
            raise self.raise_on_error(response)
        return success


class PrimitiveCallback(ResponseCallback, Generic[R]):
    @abstractmethod
    def transform(self, response: Any, **options: Any) -> Any:
        pass


class FloatCallback(PrimitiveCallback[float]):
    def transform(self, response: Any, **options: Any) -> float:
        return response if isinstance(response, float) else float(response)


class BoolCallback(PrimitiveCallback[bool]):
    def transform(self, response: Any, **options: Any) -> bool:
        if isinstance(response, bool):
            return response
        return bool(response)


class SimpleStringOrIntCallback(ResponseCallback):
    def transform(self, response: Any, **options: Any) -> Union[bool, int]:
        if isinstance(response, (int, bool)):
            return response
        else:
            return SimpleStringCallback()(response)


class TupleCallback(PrimitiveCallback[Tuple]):
    def transform(self, response: Any, **options: Any) -> Tuple:
        return tuple(response)


class ListCallback(PrimitiveCallback[List]):
    def transform(self, response: Any, **options: Any) -> List:
        if isinstance(response, list):
            return response
        return list(response)


class DateTimeCallback(ResponseCallback):
    def transform(self, response: Any, **kwargs: Any) -> datetime.datetime:
        ts = response
        if kwargs.get("unit") == "milliseconds":
            ts = ts / 1000.0
        return datetime.datetime.fromtimestamp(ts)


class DictCallback(PrimitiveCallback[Mapping]):
    def __init__(
        self,
        transform_function: Optional[Callable[[Any], Mapping]] = None,
    ):
        self.transform_function = transform_function

    def transform(self, response: Any, **options: Any) -> Mapping:
        return (
            (response if isinstance(response, dict) else dict(response))
            if not self.transform_function
            else self.transform_function(response)
        )


class SetCallback(PrimitiveCallback[Set]):
    def transform(self, response: Any, **options: Any) -> Set:
        if isinstance(response, set):
            return response
        return set(response) if response else set()


class BoolsCallback(ResponseCallback):
    def transform(self, response: Any, **options: Any) -> Tuple[bool, ...]:
        return tuple(BoolCallback()(r) for r in response)


class OptionalPrimitiveCallback(ResponseCallback, Generic[R]):
    def transform(self, response: Any, **options: Any) -> Optional[R]:
        return response


class OptionalFloatCallback(OptionalPrimitiveCallback[float]):
    def transform(self, response: Any, **options: Any) -> Optional[float]:
        if isinstance(response, float):
            return response
        return response and float(response)


class OptionalIntCallback(OptionalPrimitiveCallback[int]):
    def transform(self, response: Any, **options: Any) -> Optional[int]:
        if isinstance(response, int):
            return response
        return response and int(response)


class OptionalSetCallback(OptionalPrimitiveCallback[Set]):
    def transform(self, response: Any, **options: Any) -> Optional[Set]:
        if isinstance(response, set):
            return response
        return response and set(response)


class OptionalTupleCallback(OptionalPrimitiveCallback[Tuple]):
    def transform(self, response: Any, **options: Any) -> Optional[Tuple]:
        return response and tuple(response)



"""
coredis
-------

coredis is an async redis client with support for redis server,
cluster & sentinel.
"""

from __future__ import annotations

from coredis.client import Redis, RedisCluster
from coredis.connection import (
    BaseConnection,
    ClusterConnection,
    Connection,
    UnixDomainSocketConnection,
)
from coredis.exceptions import (
    AskError,
    AuthenticationError,
    AuthenticationFailureError,
    AuthenticationRequiredError,
    AuthorizationError,
    BusyLoadingError,
    ClusterCrossSlotError,
    ClusterDownError,
    ClusterError,
    ClusterResponseError,
    ClusterTransactionError,
    CommandNotSupportedError,
    CommandSyntaxError,
    ConnectionError,
    DataError,
    ExecAbortError,
    FunctionError,
    InvalidResponse,
    LockError,
    MovedError,
    NoKeyError,
    NoScriptError,
    PrimaryNotFoundError,
    ProtocolError,
    PubSubError,
    ReadOnlyError,
    RedisClusterException,
    RedisError,
    ReplicaNotFoundError,
    ResponseError,
    TimeoutError,
    TryAgainError,
    UnknownCommandError,
    WatchError,
)
from coredis.pool import BlockingConnectionPool, ClusterConnectionPool, ConnectionPool
from coredis.tokens import PureToken
from coredis.utils import NodeFlag

from . import _version

__all__ = [
    "Redis",
    "RedisCluster",
    "BaseConnection",
    "Connection",
    "UnixDomainSocketConnection",
    "ClusterConnection",
    "BlockingConnectionPool",
    "ConnectionPool",
    "ClusterConnectionPool",
    "AskError",
    "AuthenticationError",
    "AuthenticationFailureError",
    "AuthenticationRequiredError",
    "AuthorizationError",
    "BusyLoadingError",
    "CommandNotSupportedError",
    "ClusterCrossSlotError",
    "ClusterDownError",
    "ClusterError",
    "ClusterResponseError",
    "ClusterTransactionError",
    "CommandSyntaxError",
    "ConnectionError",
    "DataError",
    "ExecAbortError",
    "FunctionError",
    "InvalidResponse",
    "LockError",
    "MovedError",
    "NodeFlag",
    "NoKeyError",
    "NoScriptError",
    "PrimaryNotFoundError",
    "ProtocolError",
    "PubSubError",
    "PureToken",
    "ReadOnlyError",
    "RedisClusterException",
    "RedisError",
    "ReplicaNotFoundError",
    "ResponseError",
    "TimeoutError",
    "TryAgainError",
    "UnknownCommandError",
    "WatchError",
]

# For backward compatibility
StrictRedis = Redis
StrictRedisCluster = RedisCluster


__version__ = _version.get_versions()["version"]



from __future__ import annotations

import asyncio
import threading
import weakref
from asyncio import CancelledError
from concurrent.futures import Future

from coredis.commands import CommandName
from coredis.exceptions import RedisError
from coredis.response.types import MonitorResult
from coredis.typing import TYPE_CHECKING, Callable, Optional

if TYPE_CHECKING:
    import coredis.client
    import coredis.connection


class Monitor:
    """
    Monitor is useful for handling the ``MONITOR`` command to the redis server.

    It can be used as an infinite async iterator::

        async for command in client.monitor():
            print(command.time, command.type, command.command, command.args)

    Alternatively, each command can be fetched explicitly::

        monitor = client.monitor()
        command1 = await monitor.get_command()
        command2 = await monitor.get_command()
        monitor.stop()

    """

    def __init__(self, client: coredis.client.RedisConnection):
        self._client: weakref.ReferenceType[
            coredis.client.RedisConnection
        ] = weakref.ref(client)
        self.encoding = client.encoding
        self.connection: Optional[coredis.connection.Connection] = None
        self.monitoring = False

    @property
    def client(self) -> coredis.client.RedisConnection:
        client = self._client()
        assert client
        return client

    def __aiter__(self) -> Monitor:
        return self

    async def __anext__(self) -> MonitorResult:
        """
        Infinite iterator that streams back the next command processed by the
        monitored server.
        """
        return await self.get_command()

    async def get_command(self) -> MonitorResult:
        """
        Wait for the next command issued and return the details
        """
        await self.__start_monitor()
        assert self.connection
        response = await self.connection.read_response()
        if isinstance(response, bytes):
            response = response.decode(self.encoding)
        return MonitorResult.parse_response_string(response)

    async def stop(self):
        """
        Stop monitoring by issuing a ``RESET`` command
        and release the connection.
        """
        return await self.__stop_monitoring()

    def run_in_thread(
        self, response_handler=Callable[[MonitorResult], None], loop=None
    ) -> MonitorThread:
        """
        Runs the monitor in a :class:`MonitorThread` and invokes :paramref:`response_handler`
        for every command received.

        To stop the processing call :meth:`MonitorThread.stop` on the instance
        returned by this method.

        """
        monitor_thread = MonitorThread(
            self, loop or asyncio.get_event_loop(), response_handler
        )
        monitor_thread.start()
        return monitor_thread

    async def __connect(self):
        if self.connection is None:
            self.connection = await self.client.connection_pool.get_connection()

    async def __start_monitor(self):
        if self.monitoring:
            return
        await self.__connect()
        assert self.connection
        await self.connection.send_command(CommandName.MONITOR)
        response = await self.connection.read_response(decode=False)
        if not response == b"OK":  # noqa
            raise RedisError(f"Failed to start MONITOR {response}")
        self.monitoring = True

    async def __stop_monitoring(self):
        if self.connection:
            await self.connection.send_command(CommandName.RESET)
            response = await self.connection.read_response(decode=False)
            if not response == CommandName.RESET:  # noqa
                raise RedisError("Failed to reset connection")
        self.__reset()

    def __reset(self):
        if self.connection:
            self.connection.disconnect()
            self.client.connection_pool.release(self.connection)
        self.monitoring = False
        self.connection = None


class MonitorThread(threading.Thread):
    """
    Thread to be used to run monitor
    """

    def __init__(
        self,
        monitor: Monitor,
        loop: asyncio.events.AbstractEventLoop,
        response_handler: Callable[[MonitorResult], None],
    ):
        self._monitor = monitor
        self._loop = loop
        self._response_handler = response_handler
        self._future: Optional[Future[None]] = None
        super().__init__()

    def run(self):
        self._future = asyncio.run_coroutine_threadsafe(self._run(), self._loop)

    async def _run(self):
        try:
            async for command in self._monitor:
                self._response_handler(command)
        except CancelledError:
            await self._monitor.stop()

    def stop(self):
        if self._future:
            self._future.cancel()



from __future__ import annotations

import asyncio
import contextlib
import functools
import inspect
import textwrap
from abc import ABCMeta
from ssl import SSLContext
from typing import TYPE_CHECKING, Coroutine, Type, overload

from deprecated.sphinx import deprecated, versionadded
from packaging.version import Version

from coredis.commands import CommandGroup, CommandName, redis_command
from coredis.commands.core import CoreCommands
from coredis.commands.function import Library
from coredis.commands.key_spec import KeySpec
from coredis.commands.monitor import Monitor
from coredis.commands.script import Script
from coredis.commands.sentinel import SentinelCommands
from coredis.connection import Connection, RedisSSLContext, UnixDomainSocketConnection
from coredis.exceptions import (
    AskError,
    BusyLoadingError,
    ClusterDownError,
    ClusterError,
    ConnectionError,
    MovedError,
    RedisClusterException,
    ResponseError,
    TimeoutError,
    TryAgainError,
    WatchError,
)
from coredis.lock import Lock, LuaLock
from coredis.pool import ClusterConnectionPool, ConnectionPool
from coredis.response.callbacks import SimpleStringCallback
from coredis.tokens import PrefixToken, PureToken
from coredis.typing import (
    Any,
    AnyStr,
    AsyncGenerator,
    Callable,
    CommandArgList,
    Dict,
    Generic,
    Iterable,
    Iterator,
    KeyT,
    List,
    Literal,
    Optional,
    ParamSpec,
    Set,
    StringT,
    Tuple,
    TypeVar,
    Union,
    ValueT,
    add_runtime_checks,
)
from coredis.utils import NodeFlag, b, clusterdown_wrapper, first_key, nativestr
from coredis.validators import mutually_inclusive_parameters

P = ParamSpec("P")
R = TypeVar("R")

if TYPE_CHECKING:
    import coredis.commands.pipeline

CLUSTER_CUSTOM_IMPL_DEPRECATION_NOTICE = """Custom implementations for redis
cluster will be removed in version 3.5.0 and will be replaced with the implementations
in :class:`coredis.Redis`
"""


class ClusterMeta(ABCMeta):
    NODES_FLAGS: Dict[str, NodeFlag]
    RESPONSE_CALLBACKS: Dict[str, Callable]
    RESULT_CALLBACKS: Dict[str, Callable]
    NODE_FLAG_DOC_MAPPING = {
        NodeFlag.PRIMARIES: "all primaries",
        NodeFlag.REPLICAS: "all replicas",
        NodeFlag.RANDOM: "a random node",
        NodeFlag.ALL: "all nodes",
        NodeFlag.SLOT_ID: "a node selected by :paramref:`slot`",
    }

    def __new__(cls, name, bases, dct):
        kls = super().__new__(cls, name, bases, dct)
        methods = dict(k for k in inspect.getmembers(kls) if inspect.isfunction(k[1]))

        for name, method in methods.items():
            doc_addition = ""
            if cmd := getattr(method, "__coredis_command", None):
                if cmd.cluster.flag:
                    kls.NODES_FLAGS[cmd.command] = cmd.cluster.flag
                    aggregate_note = ""
                    if cmd.cluster.multi_node:
                        if cmd.cluster.combine:
                            aggregate_note = "and the results will be aggregated"
                        else:
                            aggregate_note = (
                                "and a mapping of nodes to results will be returned"
                            )
                    doc_addition = f"""
.. admonition:: Cluster note

   The command will be run on **{cls.NODE_FLAG_DOC_MAPPING[cmd.cluster.flag]}** {aggregate_note}
                    """
                if cmd.response_callback:
                    kls.RESPONSE_CALLBACKS[cmd.command] = cmd.response_callback
                if cmd.cluster.multi_node:
                    kls.RESULT_CALLBACKS[cmd.command] = cmd.cluster.combine or (
                        lambda r, **_: r
                    )
                else:
                    kls.RESULT_CALLBACKS[cmd.command] = lambda response, **_: list(
                        response.values()
                    ).pop()
                if cmd.readonly:
                    ConnectionPool.READONLY_COMMANDS.add(cmd.command)
            if doc_addition:

                def __w(
                    func: Callable[P, Coroutine[Any, Any, R]]
                ) -> Callable[P, Coroutine[Any, Any, R]]:
                    @functools.wraps(func)
                    async def _w(*a: P.args, **k: P.kwargs) -> R:
                        return await func(*a, **k)

                    _w.__doc__ = f"""{textwrap.dedent(method.__doc__)}
{doc_addition}
                """
                    return _w

                setattr(kls, name, __w(method))
        return kls


class RedisMeta(ABCMeta):
    RESPONSE_CALLBACKS: Dict

    def __new__(cls, name, bases, dct):
        kls = super().__new__(cls, name, bases, dct)
        methods = dict(k for k in inspect.getmembers(kls) if inspect.isfunction(k[1]))

        for name, method in methods.items():
            if cmd := getattr(method, "__coredis_command", None):
                if cmd.response_callback:
                    kls.RESPONSE_CALLBACKS[cmd.command] = cmd.response_callback
                setattr(kls, name, add_runtime_checks(method))
        return kls


class NodeProxy:
    class Capture:
        def __init__(self, method: str, nodes: List[Redis]):
            self.method = method
            self.nodes = nodes

        async def __call__(self, *args, **kwargs):
            awaitables = []
            for node in self.nodes:
                awaitables.append(getattr(node, self.method)(*args, **kwargs))
            responses = await asyncio.gather(*awaitables)
            return dict(zip(self.nodes, responses))

    def __init__(self, nodes):
        self.nodes = nodes

    def __getattribute__(self, name):
        nodes = object.__getattribute__(self, "nodes")
        capture = object.__getattribute__(self, "Capture")
        return capture(name, nodes)


class RedisConnection:
    encoding: str
    decode_responses: bool

    def __init__(
        self,
        host: Optional[str] = "localhost",
        port: Optional[int] = 6379,
        db: int = 0,
        username: Optional[str] = None,
        password: Optional[str] = None,
        stream_timeout: Optional[int] = None,
        connect_timeout: Optional[int] = None,
        connection_pool: Optional[ConnectionPool] = None,
        unix_socket_path: Optional[str] = None,
        encoding: str = "utf-8",
        decode_responses: bool = False,
        ssl: bool = False,
        ssl_context: Optional[SSLContext] = None,
        ssl_keyfile: Optional[str] = None,
        ssl_certfile: Optional[str] = None,
        ssl_cert_reqs: Optional[str] = None,
        ssl_ca_certs: Optional[str] = None,
        max_connections: Optional[int] = None,
        retry_on_timeout: bool = False,
        max_idle_time: int = 0,
        idle_check_interval: float = 1,
        client_name: Optional[str] = None,
        protocol_version: Literal[2, 3] = 2,
        verify_version: bool = True,
        **kwargs,
    ):
        if not connection_pool:
            kwargs = {
                "db": db,
                "username": username,
                "password": password,
                "encoding": encoding,
                "stream_timeout": stream_timeout,
                "connect_timeout": connect_timeout,
                "max_connections": max_connections,
                "retry_on_timeout": retry_on_timeout,
                "decode_responses": decode_responses,
                "max_idle_time": max_idle_time,
                "idle_check_interval": idle_check_interval,
                "client_name": client_name,
                "protocol_version": protocol_version,
            }
            # based on input, setup appropriate connection args

            if unix_socket_path is not None:
                kwargs.update(
                    {
                        "path": unix_socket_path,
                        "connection_class": UnixDomainSocketConnection,
                    }
                )
            else:
                # TCP specific options
                kwargs.update({"host": host, "port": port})

                if ssl_context is not None:
                    kwargs["ssl_context"] = ssl_context
                elif ssl:
                    ssl_context = RedisSSLContext(
                        ssl_keyfile, ssl_certfile, ssl_cert_reqs, ssl_ca_certs
                    ).get()
                    kwargs["ssl_context"] = ssl_context
            connection_pool = ConnectionPool(**kwargs)
        self.connection_pool = connection_pool
        self.encoding = encoding
        self.decode_responses = decode_responses
        self.protocol_version = protocol_version
        self.server_version: Optional[Version] = None
        self.verify_version = verify_version

    def __await__(self):
        async def closure():
            await self.connection_pool.initialize()
            return self

        return closure().__await__()

    def __repr__(self):
        return f"{type(self).__name__}<{repr(self.connection_pool)}>"

    def _ensure_server_version(self, version: Optional[str]):
        if not self.verify_version:
            return
        if not version:
            return
        if not self.server_version and version:
            self.server_version = Version(nativestr(version))
        elif str(self.server_version) != nativestr(version):
            raise Exception(
                f"Server version changed from {self.server_version} to {version}"
            )


class ResponseParser:
    RESPONSE_CALLBACKS: Dict = {}

    def __init__(self, **kwargs):
        self.response_callbacks: Dict[
            bytes, Callable
        ] = self.__class__.RESPONSE_CALLBACKS.copy()
        super().__init__(**kwargs)

    async def parse_response(
        self, connection: Connection, command_name: bytes, **options: Any
    ) -> Any:
        """
        Parses a response from the Redis server

        :meta private:
        """
        response = await connection.read_response(decode=options.get("decode"))
        if command_name in self.response_callbacks:
            callback = self.response_callbacks[command_name]
            response = callback(
                response, version=connection.protocol_version, **options
            )
        return response


class AbstractRedis(
    Generic[AnyStr],
    CoreCommands[AnyStr],
    SentinelCommands[AnyStr],
):
    """
    Async Redis client
    """

    server_version: Optional[Version]

    async def scan_iter(
        self,
        match: Optional[StringT] = None,
        count: Optional[int] = None,
        type_: Optional[StringT] = None,
    ):
        """
        Make an iterator using the SCAN command so that the client doesn't
        need to remember the cursor position.
        """
        cursor = None

        while cursor != 0:
            cursor, data = await self.scan(
                cursor=cursor, match=match, count=count, type_=type_
            )

            for item in data:
                yield item

    async def sscan_iter(
        self,
        key: KeyT,
        match: Optional[StringT] = None,
        count: Optional[int] = None,
    ):
        """
        Make an iterator using the SSCAN command so that the client doesn't
        need to remember the cursor position.
        """
        cursor = None

        while cursor != 0:
            cursor, data = await self.sscan(
                key, cursor=cursor, match=match, count=count
            )

            for item in data:
                yield item

    async def hscan_iter(
        self,
        key: KeyT,
        match: Optional[StringT] = None,
        count: Optional[int] = None,
    ) -> AsyncGenerator[Tuple[AnyStr, AnyStr], None]:
        """
        Make an iterator using the HSCAN command so that the client doesn't
        need to remember the cursor position.
        """
        cursor = None

        while cursor != 0:
            cursor, data = await self.hscan(
                key, cursor=cursor, match=match, count=count
            )

            for item in data.items():
                yield item

    async def zscan_iter(
        self,
        key: KeyT,
        match: Optional[StringT] = None,
        count: Optional[int] = None,
    ):
        """
        Make an iterator using the ZSCAN command so that the client doesn't
        need to remember the cursor position.
        """
        cursor = None

        while cursor != 0:
            cursor, data = await self.zscan(
                key,
                cursor=cursor,
                match=match,
                count=count,
            )

            for item in data:
                yield item

    def register_script(self, script: ValueT) -> Script:
        """
        Registers a Lua :paramref:`script`

        :return: A :class:`coredis.commands.script.Script` instance that is
         callable and hides the complexity of dealing with scripts, keys, and
         shas.
        """
        return Script(self, script)  # type: ignore

    @versionadded(version="3.1.0")
    async def register_library(
        self,
        name: StringT,
        code: StringT,
    ) -> Library:
        """
        Register a new library

        :param name: name of the library
        :param code: raw code for the library
        """
        return await Library[AnyStr](self, name=name, code=code)

    @versionadded(version="3.1.0")
    async def get_library(self, name: StringT) -> Library:
        """
        Fetch a pre registered library

        :param name: name of the library
        """
        return await Library[AnyStr](self, name)


class AbstractRedisCluster(AbstractRedis[AnyStr]):
    @deprecated(version="3.1.0", reason=CLUSTER_CUSTOM_IMPL_DEPRECATION_NOTICE)
    @redis_command(
        CommandName.RENAME,
        group=CommandGroup.GENERIC,
        response_callback=SimpleStringCallback(),
    )
    async def rename(self, key: KeyT, newkey: KeyT) -> bool:
        """
        Rename key :paramref:`key` to :paramref:`newkey`

        .. admonition:: Cluster note

           This operation is no longer atomic because each key must be queried
           then set in separate calls because they maybe will change cluster node
        """

        if key == newkey:
            raise ResponseError("source and destination objects are the same")

        data = await self.dump(key)

        if data is None:
            raise ResponseError("no such key")

        ttl = await self.pttl(key)

        if ttl is None or ttl < 1:
            ttl = 0

        await self.delete([newkey])
        await self.restore(newkey, ttl, data)
        await self.delete([key])

        return True

    @deprecated(version="3.1.0", reason=CLUSTER_CUSTOM_IMPL_DEPRECATION_NOTICE)
    @redis_command(CommandName.DEL, group=CommandGroup.GENERIC)
    async def delete(self, keys: Iterable[KeyT]) -> int:
        """
        "Delete one or more keys specified by :paramref:`keys`"
        """
        count = 0

        for arg in keys:
            count += await self.execute_command(CommandName.DEL, arg)

        return count

    @deprecated(version="3.1.0", reason=CLUSTER_CUSTOM_IMPL_DEPRECATION_NOTICE)
    @redis_command(
        CommandName.RENAMENX,
        group=CommandGroup.GENERIC,
    )
    async def renamenx(self, key: KeyT, newkey: KeyT) -> bool:
        """
        Rekeys key paramref:`key` to :paramref:`newkey` if
        :paramref:`newkey` doesn't already exist

        :return: False when :paramref:`newkey` already exists.
        """

        if not await self.exists([newkey]) == 1:
            return await self.rename(key, newkey)

        return False

    @deprecated(version="3.1.0", reason=CLUSTER_CUSTOM_IMPL_DEPRECATION_NOTICE)
    @mutually_inclusive_parameters("offset", "count")
    @redis_command(
        CommandName.SORT,
        group=CommandGroup.GENERIC,
    )
    async def sort(
        self,
        key: KeyT,
        gets: Optional[Iterable[KeyT]] = None,
        by: Optional[StringT] = None,
        offset: Optional[int] = None,
        count: Optional[int] = None,
        order: Optional[Literal[PureToken.ASC, PureToken.DESC]] = None,
        alpha: Optional[bool] = None,
        store: Optional[KeyT] = None,
    ) -> Union[Tuple[AnyStr, ...], int]:
        """
        Sort the elements in a list, set or sorted set

        :return: sorted elements.

         When the :paramref:`store` option is specified the command returns
         the number of sorted elements in the destination list.
        """

        try:
            data_type = nativestr(await self.type(key))

            if data_type == "none":
                return ()
            elif data_type == "set":
                data = list(await self.smembers(key))[:]
            elif data_type == "list":
                data = await self.lrange(key, 0, -1)
            else:
                raise RedisClusterException(f"Unable to sort data type : {data_type}")

            if by is not None:
                # _sort_using_by_arg mutates data so we don't
                # need need a return value.
                data = await self._sort_using_by_arg(data, by, alpha)
            elif not alpha:
                data.sort(key=self._strtod_key_func)
            else:
                data.sort()

            if order == PureToken.DESC:
                data = data[::-1]

            if not (offset is None or count is None):
                data = data[offset : offset + count]

            if gets:
                data = await self._retrive_data_from_sort(data, gets)

            if store is not None:
                if data_type == "set":
                    await self.delete([store])
                    await self.rpush(store, data)
                elif data_type == "list":
                    await self.delete([store])
                    await self.rpush(store, data)
                else:
                    raise RedisClusterException(
                        f"Unable to store sorted data for data type : {data_type}"
                    )

                return len(data)

            return tuple(data)
        except KeyError:
            return ()

    async def _retrive_data_from_sort(self, data, gets) -> List[AnyStr]:
        """
        Used by sort()
        """

        if gets:
            new_data: List[AnyStr] = []

            for k in data:
                for g in gets:
                    single_item = await self._get_single_item(k, g)
                    if single_item:
                        new_data.append(single_item)
            data = new_data

        return data

    async def _get_single_item(self, k, g) -> Optional[AnyStr]:
        """
        Used by sort()
        """

        if "*" in g:
            g = g.replace("*", k)

            if "->" in g:
                key, hash_key = g.split("->")
                single_item = (await self.hgetall(key) or {}).get(hash_key)
            else:
                single_item = await self.get(g)
        elif "#" in g:
            single_item = k
        else:
            single_item = None

        return single_item

    def _strtod_key_func(self, arg):
        """
        Used by sort()
        """

        return float(arg)

    async def _sort_using_by_arg(self, data, by, alpha):
        """
        Used by sort()
        """

        async def _by_key(arg):
            key = by.replace("*", arg)

            if "->" in by:
                key, hash_key = key.split("->")
                v = await self.hget(key, hash_key)

                if v is not None:
                    if alpha:
                        return v
                    else:
                        return float(v)
            else:
                return await self.get(key)

        sorted_data = []

        for d in data:
            sorted_data.append((d, await _by_key(d)))

        return [x[0] for x in sorted(sorted_data, key=lambda x: x[1])]

    @deprecated(version="3.1.0", reason=CLUSTER_CUSTOM_IMPL_DEPRECATION_NOTICE)
    @redis_command(CommandName.MGET, readonly=True, group=CommandGroup.STRING)
    async def mget(self, keys: Iterable[KeyT]) -> Tuple[Optional[AnyStr], ...]:
        """
        Returns values ordered identically to :paramref:`keys`

        .. admonition:: Cluster note

           Iterate all keys and send GET for each key.
           This will go alot slower than a normal mget call in Redis.
           **Operation is no longer atomic**
        """
        res = list()

        for arg in keys:
            res.append(await self.get(arg))

        return tuple(res)

    @deprecated(version="3.1.0", reason=CLUSTER_CUSTOM_IMPL_DEPRECATION_NOTICE)
    @redis_command(
        CommandName.MSET,
        group=CommandGroup.STRING,
    )
    async def mset(self, key_values: Dict[KeyT, ValueT]) -> bool:
        """
        Sets key/values based on a mapping. Mapping can be supplied as a single
        dictionary argument or as kwargs.

        .. admonition:: Cluster note

           Iterate over all items and do SET on each (k,v) pair.
           **Operation is no longer atomic**
        """

        for pair in iter(key_values.items()):
            await self.set(pair[0], pair[1])

        return True

    @deprecated(version="3.1.0", reason=CLUSTER_CUSTOM_IMPL_DEPRECATION_NOTICE)
    @redis_command(
        CommandName.MSETNX,
        group=CommandGroup.STRING,
    )
    async def msetnx(self, key_values: Dict[KeyT, ValueT]) -> bool:
        """
        Sets key/values based on a mapping if none of the keys are already set.
        Mapping can be supplied as a single dictionary argument or as kwargs.
        Returns a boolean indicating if the operation was successful.

        .. admonition:: Cluster note

            Iterate over all items and do GET to determine if all keys do not exists.
            If true then call mset() on all keysu.
            **Operation is no longer atomic**
        """

        for k, _ in key_values.items():
            if await self.get(k):
                return False

        return await self.mset(key_values)

    @deprecated(version="3.1.0", reason=CLUSTER_CUSTOM_IMPL_DEPRECATION_NOTICE)
    @redis_command(
        CommandName.SDIFF,
        readonly=True,
        group=CommandGroup.SET,
    )
    async def sdiff(self, keys: Iterable[KeyT]) -> Set[AnyStr]:
        """
        Returns the difference of sets specified by :paramref:`keys`

        .. admonition:: Cluster note

           Query all keys and diff all sets in memory and return the result
        """
        _keys = list(keys)
        res = await self.smembers(_keys[0])

        for arg in _keys[1:]:
            res -= await self.smembers(arg)

        return res

    @deprecated(version="3.1.0", reason=CLUSTER_CUSTOM_IMPL_DEPRECATION_NOTICE)
    @redis_command(
        CommandName.SDIFFSTORE,
        group=CommandGroup.SET,
    )
    async def sdiffstore(self, keys: Iterable[KeyT], destination: KeyT) -> int:
        """
        Stores the difference of sets specified by :paramref:`keys` into a new
        set named :paramref:`destination`.  Returns the number of keys in the new set.
        Overwrites dest key if it exists.

        .. admonition:: Cluster note

           Use sdiff() --> Delete dest key --> store result in dest key
        """
        res = await self.sdiff(keys)
        await self.delete([destination])

        if not res:
            return 0

        return await self.sadd(destination, res)

    @deprecated(version="3.1.0", reason=CLUSTER_CUSTOM_IMPL_DEPRECATION_NOTICE)
    @redis_command(
        CommandName.SINTER,
        readonly=True,
        group=CommandGroup.SET,
    )
    async def sinter(self, keys: Iterable[KeyT]) -> Set[AnyStr]:
        """
        Returns the intersection of sets specified by :paramref:`keys`

        .. admonition:: Cluster note

           Query all keys, perform intersection in memory and return the result
        """
        _keys = list(keys)
        res = await self.smembers(_keys[0])

        for arg in _keys[1:]:
            res &= await self.smembers(arg)

        return res

    @deprecated(version="3.1.0", reason=CLUSTER_CUSTOM_IMPL_DEPRECATION_NOTICE)
    @redis_command(
        CommandName.SINTERSTORE,
        group=CommandGroup.SET,
    )
    async def sinterstore(self, keys: Iterable[KeyT], destination: KeyT) -> int:
        """
        Stores the intersection of sets specified by :paramref:`keys` into a new
        set named :paramref:`destination`.  Returns the number of keys in the new set.

        .. admonition:: Cluster note

           Uses sinter() --> Delete dest key --> store result in dest key
        """
        res = await self.sinter(keys)
        await self.delete([destination])

        if res:
            await self.sadd(destination, res)

            return len(res)
        else:
            return 0

    @deprecated(version="3.1.0", reason=CLUSTER_CUSTOM_IMPL_DEPRECATION_NOTICE)
    @redis_command(
        CommandName.SMOVE,
        group=CommandGroup.SET,
    )
    async def smove(self, source: KeyT, destination: KeyT, member: ValueT) -> bool:
        """
        Moves :paramref:`member` from set :paramref:`source` to set :paramref:`destination`

        .. admonition:: Cluster note

           Uses smembers() --> srem() --> sadd.
           **Function is no longer atomic**
        """
        res = await self.srem(source, [member])

        # Only add the element if existed in src set

        if res == 1:
            await self.sadd(destination, [member])

        return bool(res)

    @deprecated(version="3.1.0", reason=CLUSTER_CUSTOM_IMPL_DEPRECATION_NOTICE)
    @redis_command(
        CommandName.SUNION,
        readonly=True,
        group=CommandGroup.SET,
    )
    async def sunion(self, keys: Iterable[KeyT]) -> Set[AnyStr]:
        """
        Returns the union of sets specified by :paramref:`keys`

        .. admonition:: Cluster note

           Query all keys, perform union in memory and return result

           **Operation is no longer atomic**
        """
        _keys = list(keys)
        res = await self.smembers(_keys[0])

        for arg in _keys[1:]:
            res |= await self.smembers(arg)

        return res

    @deprecated(version="3.1.0", reason=CLUSTER_CUSTOM_IMPL_DEPRECATION_NOTICE)
    @redis_command(
        CommandName.SUNIONSTORE,
        group=CommandGroup.SET,
    )
    async def sunionstore(self, keys: Iterable[KeyT], destination: KeyT) -> int:
        """
        Stores the union of sets specified by :paramref:`keys` into a new
        set named :paramref:`destination`.  Returns the number of keys in the new set.

        .. admonition:: Cluster note

           Use sunion() --> delete() dest key --> store result in dest key

           **Operation is no longer atomic**
        """
        res = await self.sunion(keys)
        await self.delete([destination])

        return await self.sadd(destination, res)

    @deprecated(version="3.1.0", reason=CLUSTER_CUSTOM_IMPL_DEPRECATION_NOTICE)
    @redis_command(
        CommandName.BRPOPLPUSH,
        version_deprecated="6.2.0",
        group=CommandGroup.LIST,
    )
    async def brpoplpush(
        self, source: KeyT, destination: KeyT, timeout: Union[int, float]
    ) -> Optional[AnyStr]:
        """
        Pops a value off the tail of :paramref:`source`, push it on the head
        of :paramref:`destination` and then return it.

        This command blocks until a value is in :paramref:`source` or until
        :paramref:`timeout` seconds elapse, whichever is first. A :paramref:`timeout`
        value of 0 blocks forever.

        .. admonition:: Cluster note

           Calls brpop() then sends the result into lpush()

           **Operation is no longer atomic**
        """
        try:
            values = await self.brpop([source], timeout=timeout)

            if not values:
                return None
        except TimeoutError:
            return None
        await self.lpush(destination, [values[1]])

        return values[1]

    @deprecated(version="3.1.0", reason=CLUSTER_CUSTOM_IMPL_DEPRECATION_NOTICE)
    @redis_command(
        CommandName.RPOPLPUSH,
        version_deprecated="6.2.0",
        group=CommandGroup.LIST,
    )
    async def rpoplpush(self, source: KeyT, destination: KeyT) -> Optional[AnyStr]:
        """
        RPOP a value off of the :paramref:`source` list and atomically LPUSH it
        on to the :paramref:`destination` list.  Returns the value.

        .. admonition:: Cluster note

           Calls :meth:`rpop` then send the result into :meth:`lpush`

           **Operation is no longer atomic**
        """
        value = await self.rpop(source)

        if value:
            await self.lpush(destination, [value])
            return value
        return None


class Redis(
    AbstractRedis[AnyStr],
    Generic[AnyStr],
    ResponseParser,
    RedisConnection,
    metaclass=RedisMeta,
):
    @overload
    def __init__(
        self: Redis[bytes],
        host: Optional[str] = ...,
        port: Optional[int] = ...,
        db: int = ...,
        *,
        username: Optional[str] = ...,
        password: Optional[str] = ...,
        stream_timeout: Optional[int] = ...,
        connect_timeout: Optional[int] = ...,
        connection_pool: Optional[ConnectionPool] = ...,
        unix_socket_path: Optional[str] = ...,
        encoding: str = ...,
        decode_responses: Literal[False] = ...,
        ssl: bool = ...,
        ssl_context: Optional[SSLContext] = ...,
        ssl_keyfile: Optional[str] = ...,
        ssl_certfile: Optional[str] = ...,
        ssl_cert_reqs: Optional[str] = ...,
        ssl_ca_certs: Optional[str] = ...,
        max_connections: Optional[int] = ...,
        retry_on_timeout: bool = ...,
        max_idle_time: float = ...,
        idle_check_interval: float = ...,
        client_name: Optional[str] = ...,
        protocol_version: Literal[2, 3] = 2,
        verify_version: bool = ...,
        **_,
    ):
        ...

    @overload
    def __init__(
        self: Redis[str],
        host: Optional[str] = ...,
        port: Optional[int] = ...,
        db: int = ...,
        *,
        username: Optional[str] = ...,
        password: Optional[str] = ...,
        stream_timeout: Optional[int] = ...,
        connect_timeout: Optional[int] = ...,
        connection_pool: Optional[ConnectionPool] = ...,
        unix_socket_path: Optional[str] = ...,
        encoding: str = ...,
        decode_responses: Literal[True],
        ssl: bool = ...,
        ssl_context: Optional[SSLContext] = ...,
        ssl_keyfile: Optional[str] = ...,
        ssl_certfile: Optional[str] = ...,
        ssl_cert_reqs: Optional[str] = ...,
        ssl_ca_certs: Optional[str] = ...,
        max_connections: Optional[int] = ...,
        retry_on_timeout: bool = ...,
        max_idle_time: float = ...,
        idle_check_interval: float = ...,
        client_name: Optional[str] = ...,
        protocol_version: Literal[2, 3] = 2,
        verify_version: bool = ...,
        **_,
    ):
        ...

    def __init__(
        self,
        host: Optional[str] = "localhost",
        port: Optional[int] = 6379,
        db: int = 0,
        *,
        username: Optional[str] = None,
        password: Optional[str] = None,
        stream_timeout: Optional[int] = None,
        connect_timeout: Optional[int] = None,
        connection_pool: Optional[ConnectionPool] = None,
        unix_socket_path: Optional[str] = None,
        encoding: str = "utf-8",
        decode_responses: bool = False,
        ssl: bool = False,
        ssl_context: Optional[SSLContext] = None,
        ssl_keyfile: Optional[str] = None,
        ssl_certfile: Optional[str] = None,
        ssl_cert_reqs: Optional[str] = None,
        ssl_ca_certs: Optional[str] = None,
        max_connections: Optional[int] = None,
        retry_on_timeout: bool = False,
        max_idle_time: float = 0,
        idle_check_interval: float = 1,
        client_name: Optional[str] = None,
        protocol_version: Literal[2, 3] = 2,
        verify_version: bool = False,
        **_,
    ):
        """
        Initializes a new Redis client
        """
        super().__init__(
            host=host,
            port=port,
            db=db,
            username=username,
            password=password,
            stream_timeout=stream_timeout,
            connect_timeout=connect_timeout,
            connection_pool=connection_pool,
            unix_socket_path=unix_socket_path,
            encoding=encoding,
            decode_responses=decode_responses,
            ssl=ssl,
            ssl_context=ssl_context,
            ssl_keyfile=ssl_keyfile,
            ssl_certfile=ssl_certfile,
            ssl_cert_reqs=ssl_cert_reqs,
            ssl_ca_certs=ssl_ca_certs,
            max_connections=max_connections,
            retry_on_timeout=retry_on_timeout,
            max_idle_time=max_idle_time,
            idle_check_interval=idle_check_interval,
            client_name=client_name,
            protocol_version=protocol_version,
            verify_version=verify_version,
        )
        self._use_lua_lock: Optional[bool] = None

    @classmethod
    @overload
    def from_url(
        cls,
        url: str,
        db: Optional[int] = ...,
        *,
        decode_responses: Literal[False] = ...,
        **kwargs,
    ) -> Redis[bytes]:
        ...

    @classmethod
    @overload
    def from_url(
        cls,
        url: str,
        db: Optional[int] = ...,
        *,
        decode_responses: Literal[True],
        **kwargs,
    ) -> Redis[str]:
        ...

    @classmethod
    def from_url(
        cls,
        url: str,
        db: Optional[int] = None,
        *,
        decode_responses: bool = False,
        **kwargs,
    ):
        """
        Return a Redis client object configured from the given URL, which must
        use either the `redis:// scheme
        <http://www.iana.org/assignments/uri-schemes/prov/redis>`_ for RESP
        connections or the ``unix://`` scheme for Unix domain sockets.

        For example:

        - ``redis://[:password]@localhost:6379/0``
        - ``unix://[:password]@/path/to/socket.sock?db=0``

        There are several ways to specify a database number. The parse function
        will return the first specified option:

        1. A ``db`` querystring option, e.g. ``redis://localhost?db=0``
        2. If using the redis:// scheme, the path argument of the url, e.g.
           ``redis://localhost/0``
        3. The ``db`` argument to this function.

        If none of these options are specified, ``db=0`` is used.


        Any additional querystring arguments and keyword arguments will be
        passed along to the :class:`ConnectionPool` initializer. In the case
        of conflicting arguments, querystring arguments always win.
        """
        connection_pool: ConnectionPool = ConnectionPool.from_url(
            url, db=db, decode_responses=decode_responses, **kwargs
        )

        if decode_responses:
            return Redis[str](decode_responses=True, connection_pool=connection_pool)
        else:
            return Redis[bytes](decode_responses=False, connection_pool=connection_pool)

    def set_response_callback(self, command, callback):
        """
        Sets a custom Response Callback

        :meta private:
        """
        self.response_callbacks[command] = callback

    async def execute_command(
        self, command: StringT, *args: Any, **options: Any
    ) -> Any:
        """Executes a command and returns a parsed response"""
        pool = self.connection_pool
        connection = await pool.get_connection()
        self._ensure_server_version(connection.server_version)

        try:
            await connection.send_command(command, *args)

            return await self.parse_response(connection, b(command), **options)
        except asyncio.CancelledError:
            # do not retry when coroutine is cancelled
            connection.disconnect()
            raise
        except (ConnectionError, TimeoutError) as e:
            connection.disconnect()

            if not connection.retry_on_timeout and isinstance(e, TimeoutError):
                raise
            await connection.send_command(command, *args)

            return await self.parse_response(connection, b(command), **options)
        finally:
            pool.release(connection)

    def monitor(self) -> Monitor:
        """
        Return an instance of a :class:`~coredis.commands.monitor.Monitor`

        The monitor can be used as an async iterator or individual commands
        can be fetched via :meth:`~coredis.commands.monitor.Monitor.get_command`.
        """
        return Monitor(self)

    def lock(
        self,
        name: StringT,
        timeout: Optional[float] = None,
        sleep: float = 0.1,
        blocking_timeout: Optional[bool] = None,
        lock_class: Optional[Type[Lock]] = None,
        thread_local: bool = True,
    ) -> Lock:
        """
        Return a new :class:`~coredis.lock.Lock` object using :paramref:`name` that mimics
        the behavior of :class:`threading.Lock`.

        See: :class:`~coredis.lock.LuaLock` (the default :paramref:`lock_class`)
        for more details.

        :raises: :exc:`~coredis.LockError`
        """

        if lock_class is None:
            if self._use_lua_lock is None:  # type: ignore
                # the first time .lock() is called, determine if we can use
                # Lua by attempting to register the necessary scripts
                try:
                    LuaLock.register_scripts(self)
                    self._use_lua_lock = True
                except ResponseError:
                    self._use_lua_lock = False
            lock_class = self._use_lua_lock and LuaLock or Lock

        return lock_class(
            self,
            name,
            timeout=timeout,
            sleep=sleep,
            blocking_timeout=blocking_timeout,
            thread_local=thread_local,
        )

    def pubsub(self, ignore_subscribe_messages: bool = False, **kwargs):
        """
        Return a Publish/Subscribe object. With this object, you can
        subscribe to channels and listen for messages that get published to
        them.
        """
        from coredis.commands.pubsub import PubSub

        return PubSub(
            self.connection_pool,
            ignore_subscribe_messages=ignore_subscribe_messages,
            **kwargs,
        )

    async def pipeline(
        self,
        transaction: Optional[bool] = True,
        watches: Optional[Iterable[StringT]] = None,
    ) -> "coredis.commands.pipeline.Pipeline[AnyStr]":
        """
        Returns a new pipeline object that can queue multiple commands for
        later execution. :paramref:`transaction` indicates whether all commands
        should be executed atomically. Apart from making a group of operations
        atomic, pipelines are useful for reducing the back-and-forth overhead
        between the client and server.
        """
        from coredis.commands.pipeline import Pipeline

        pipeline: Pipeline[AnyStr] = Pipeline.proxy(
            self.connection_pool, self.response_callbacks, transaction
        )
        await pipeline.reset()
        return pipeline

    async def transaction(self, func, *watches: StringT, **kwargs):
        """
        Convenience method for executing the callable :paramref:`func` as a
        transaction while watching all keys specified in :paramref:`watches`.
        The :paramref:`func` callable should expect a single argument which is a
        :class:`coredis.commands.pipeline.Pipeline` object retrieved by calling
        :meth:`~coredis.Redis.pipeline`.
        """
        value_from_callable = kwargs.pop("value_from_callable", False)
        watch_delay = kwargs.pop("watch_delay", None)
        async with await self.pipeline(True) as pipe:
            while True:
                try:
                    if watches:
                        await pipe.watch(*watches)
                    func_value = await func(pipe)
                    exec_value = await pipe.execute()
                    return func_value if value_from_callable else exec_value
                except WatchError:
                    if watch_delay is not None and watch_delay > 0:
                        await asyncio.sleep(watch_delay)
                    continue


class RedisCluster(
    AbstractRedisCluster[AnyStr], ResponseParser, RedisConnection, metaclass=ClusterMeta
):

    RedisClusterRequestTTL = 16
    NODES_FLAGS: Dict[bytes, NodeFlag] = {}
    RESPONSE_CALLBACKS: Dict[bytes, Callable] = {}
    RESULT_CALLBACKS: Dict[bytes, Callable] = {}

    connection_pool: ClusterConnectionPool

    @overload
    def __init__(
        self: RedisCluster[bytes],
        host: Optional[str] = ...,
        port: Optional[int] = ...,
        *,
        startup_nodes: Optional[Iterable[Dict]] = ...,
        max_connections: int = ...,
        max_connections_per_node: bool = ...,
        readonly: bool = ...,
        reinitialize_steps: Optional[int] = ...,
        skip_full_coverage_check: bool = ...,
        nodemanager_follow_cluster: bool = ...,
        decode_responses: Literal[False] = ...,
        connection_pool: Optional[ClusterConnectionPool] = ...,
        verify_version: bool = ...,
        **kwargs,
    ):
        ...

    @overload
    def __init__(
        self: RedisCluster[str],
        host: Optional[str] = ...,
        port: Optional[int] = ...,
        *,
        startup_nodes: Optional[Iterable[Dict]] = ...,
        max_connections: int = ...,
        max_connections_per_node: bool = ...,
        readonly: bool = ...,
        reinitialize_steps: Optional[int] = ...,
        skip_full_coverage_check: bool = ...,
        nodemanager_follow_cluster: bool = ...,
        decode_responses: Literal[True],
        connection_pool: Optional[ClusterConnectionPool] = ...,
        verify_version: bool = ...,
        **kwargs,
    ):
        ...

    def __init__(
        self,
        host: Optional[str] = None,
        port: Optional[int] = None,
        *,
        startup_nodes: Optional[Iterable[Dict]] = None,
        max_connections: int = 32,
        max_connections_per_node: bool = False,
        readonly: bool = False,
        reinitialize_steps: Optional[int] = None,
        skip_full_coverage_check: bool = False,
        nodemanager_follow_cluster: bool = False,
        decode_responses: bool = False,
        connection_pool: Optional[ClusterConnectionPool] = None,
        verify_version: bool = False,
        **kwargs,
    ):
        """
        :param host: Can be used to point to a startup node
        :param port: Can be used to point to a startup node
        :param startup_nodes: List of nodes that initial bootstrapping can be done
         from
        :param max_connections: Maximum number of connections that should be kept open at one time
        :param readonly: enable READONLY mode. You can read possibly stale data from slave.
        :param skip_full_coverage_check: Skips the check of cluster-require-full-coverage config,
         useful for clusters without the CONFIG command (like aws)
        :param nodemanager_follow_cluster: The node manager will during initialization try the
         last set of nodes that it was operating on. This will allow the client to drift along
         side the cluster if the cluster nodes move around alot.
        :param verify_version: Ensure requested methods are supported on the target server
        """

        if "db" in kwargs:
            raise RedisClusterException(
                "Argument 'db' is not possible to use in cluster mode"
            )

        if connection_pool:
            pool = connection_pool
        else:
            startup_nodes = [] if startup_nodes is None else list(startup_nodes)

            # Support host/port as argument

            if host:
                startup_nodes.append({"host": host, "port": port if port else 7000})
            pool = ClusterConnectionPool(
                startup_nodes=startup_nodes,
                max_connections=max_connections,
                reinitialize_steps=reinitialize_steps,
                max_connections_per_node=max_connections_per_node,
                skip_full_coverage_check=skip_full_coverage_check,
                nodemanager_follow_cluster=nodemanager_follow_cluster,
                readonly=readonly,
                decode_responses=decode_responses,
                **kwargs,
            )

        super().__init__(
            connection_pool=pool,
            decode_responses=decode_responses,
            verify_version=verify_version,
            **kwargs,
        )

        self.refresh_table_asap: bool = False
        self.nodes_flags: Dict[bytes, NodeFlag] = self.__class__.NODES_FLAGS.copy()
        self.result_callbacks: Dict[
            bytes, Callable
        ] = self.__class__.RESULT_CALLBACKS.copy()
        self.response_callbacks: Dict[
            bytes, Callable
        ] = self.__class__.RESPONSE_CALLBACKS.copy()

    @classmethod
    @overload
    def from_url(
        cls,
        url: str,
        *,
        db: Optional[int] = ...,
        skip_full_coverage_check: bool = ...,
        decode_responses: Literal[False] = ...,
        **kwargs,
    ) -> RedisCluster[bytes]:
        ...

    @classmethod
    @overload
    def from_url(
        cls,
        url: str,
        *,
        db: Optional[int] = ...,
        skip_full_coverage_check: bool = ...,
        decode_responses: Literal[True],
        **kwargs,
    ) -> RedisCluster[str]:
        ...

    @classmethod
    def from_url(
        cls,
        url: str,
        *,
        db: Optional[int] = None,
        skip_full_coverage_check: bool = False,
        decode_responses: bool = False,
        **kwargs,
    ):
        """
        Return a Redis client object configured from the given URL, which must
        use either the ``redis://`` scheme
        `<http://www.iana.org/assignments/uri-schemes/prov/redis>`_ for RESP
        connections or the ``unix://`` scheme for Unix domain sockets.
        For example:

            - ``redis://[:password]@localhost:6379/0``
            - ``unix://[:password]@/path/to/socket.sock?db=0``

        There are several ways to specify a database number. The parse function
        will return the first specified option:

            #. A ``db`` querystring option, e.g. ``redis://localhost?db=0``
            #. If using the ``redis://`` scheme, the path argument of the url, e.g.
               ``redis://localhost/0``
            #. The ``db`` argument to this function.

        If none of these options are specified, db=0 is used.

        Any additional querystring arguments and keyword arguments will be
        passed along to the :class:`ConnectionPool` class's initializer. In the case
        of conflicting arguments, querystring arguments always win.
        """
        connection_pool = ClusterConnectionPool.from_url(
            url,
            db=db,
            skip_full_coverage_check=skip_full_coverage_check,
            decode_responses=decode_responses,
            **kwargs,
        )
        if decode_responses:
            return RedisCluster[str](
                decode_responses=True, connection_pool=connection_pool
            )
        else:
            return RedisCluster[bytes](
                decode_responses=False, connection_pool=connection_pool
            )

    def __repr__(self):
        servers = list(
            {
                "{}:{}".format(info["host"], info["port"])
                for info in self.connection_pool.nodes.startup_nodes
            }
        )
        servers.sort()

        return "{}<{}>".format(type(self).__name__, ", ".join(servers))

    @property
    def all_nodes(self) -> Iterator[Redis]:
        """ """
        for node in self.connection_pool.nodes.all_nodes():
            yield self.connection_pool.nodes.get_redis_link(node["host"], node["port"])

    @property
    def primaries(self) -> Iterator[Redis]:
        """ """
        for master in self.connection_pool.nodes.all_primaries():
            yield self.connection_pool.nodes.get_redis_link(
                master["host"], master["port"]
            )

    @property
    def replicas(self) -> Iterator[Redis]:
        """ """
        for replica in self.connection_pool.nodes.all_replicas():
            yield self.connection_pool.nodes.get_redis_link(
                replica["host"], replica["port"]
            )

    def set_result_callback(self, command, callback):
        "Sets a custom Result Callback"
        self.result_callbacks[command] = callback

    def _determine_slot(self, *args):
        """Figures out what slot based on command and args"""

        if len(args) <= 1:
            raise RedisClusterException(
                f"No way to dispatch this command:{args} to Redis Cluster. Missing key."
            )
        command: bytes = args[0]
        keys: Tuple[ValueT, ...] = KeySpec.extract_keys(args)

        if (
            command in {CommandName.EVAL, CommandName.EVALSHA, CommandName.FCALL}
            and not keys
        ):
            return

        slots = {self.connection_pool.nodes.keyslot(key) for key in keys}
        if not slots:
            raise RedisClusterException(
                f"No way to dispatch this command: {args} to Redis Cluster."
            )

        if len(slots) != 1:
            raise RedisClusterException(
                f"{str(command)} - all keys must map to the same key slot"
            )
        return slots.pop()

    def _merge_result(self, command, res, **kwargs):
        """
        `res` is a dict with the following structure Dict(NodeName, CommandResult)
        """

        if command in self.result_callbacks:
            return self.result_callbacks[command](res, **kwargs)

        # Default way to handle result

        return first_key(res)

    def determine_node(self, *args, **kwargs):
        """
        TODO: document
        """
        command = args[0]
        node_flag = self.nodes_flags.get(command)

        if node_flag == NodeFlag.BLOCKED:
            raise RedisClusterException(
                f"Command: {command} is blocked in redis cluster mode"
            )
        elif node_flag == NodeFlag.RANDOM:
            return [self.connection_pool.nodes.random_node()]
        elif node_flag == NodeFlag.PRIMARIES:
            return self.connection_pool.nodes.all_primaries()
        elif node_flag == NodeFlag.ALL:
            return self.connection_pool.nodes.all_nodes()
        elif node_flag == NodeFlag.SLOT_ID:
            # if node flag of command is SLOT_ID
            # `slot_id` should is assumed in kwargs
            slot = kwargs.get("slot_id")

            if not slot:
                raise RedisClusterException(
                    f"slot_id is needed to execute command {command}"
                )

            return [self.connection_pool.nodes.node_from_slot(slot)]
        else:
            return None

    @clusterdown_wrapper
    async def execute_command(self, command: StringT, *args: Any, **kwargs: Any):
        """
        Sends a command to a node in the cluster
        """
        if not self.connection_pool.initialized:
            await self.connection_pool.initialize()

        if not command:
            raise RedisClusterException("Unable to determine command to use")

        node = self.determine_node(command, *args, **kwargs)

        if node:
            return await self.execute_command_on_nodes(node, command, *args, **kwargs)

        # If set externally we must update it before calling any commands

        if self.refresh_table_asap:
            await self.connection_pool.nodes.initialize()
            self.refresh_table_asap = False

        redirect_addr = None
        asking = False

        try_random_node = False
        try_random_type = NodeFlag.ALL
        slot = self._determine_slot(command, *args)
        if not slot:
            try_random_node = True
            try_random_type = NodeFlag.PRIMARIES

        ttl = int(self.RedisClusterRequestTTL)

        while ttl > 0:
            ttl -= 1

            if asking:
                node = self.connection_pool.nodes.nodes[redirect_addr]
                r = self.connection_pool.get_connection_by_node(node)
            elif try_random_node:
                r = self.connection_pool.get_random_connection(
                    primary=try_random_type == NodeFlag.PRIMARIES
                )
                try_random_node = False
            else:
                if self.refresh_table_asap:
                    # MOVED
                    node = self.connection_pool.get_master_node_by_slot(slot)
                else:
                    node = self.connection_pool.get_node_by_slot(slot, command)
                r = self.connection_pool.get_connection_by_node(node)
            self._ensure_server_version(r.server_version)

            try:
                if asking:
                    await r.send_command(CommandName.ASKING)
                    await self.parse_response(r, CommandName.ASKING, **kwargs)
                    asking = False

                await r.send_command(command, *args)

                return await self.parse_response(r, b(command), **kwargs)
            except (RedisClusterException, BusyLoadingError, asyncio.CancelledError):
                raise
            except (ConnectionError, TimeoutError):
                try_random_node = True

                if ttl < self.RedisClusterRequestTTL / 2:
                    await asyncio.sleep(0.1)
            except ClusterDownError as e:
                self.connection_pool.disconnect()
                self.connection_pool.reset()
                self.refresh_table_asap = True

                raise e
            except MovedError as e:
                # Reinitialize on ever x number of MovedError.
                # This counter will increase faster when the same client object
                # is shared between multiple threads. To reduce the frequency you
                # can set the variable 'reinitialize_steps' in the constructor.
                self.refresh_table_asap = True
                await self.connection_pool.nodes.increment_reinitialize_counter()

                node = self.connection_pool.nodes.set_node(
                    e.host, e.port, server_type="master"
                )
                self.connection_pool.nodes.slots[e.slot_id][0] = node
            except TryAgainError:
                if ttl < self.RedisClusterRequestTTL / 2:
                    await asyncio.sleep(0.05)
            except AskError as e:
                redirect_addr, asking = f"{e.host}:{e.port}", True
            finally:

                self.connection_pool.release(r)

        raise ClusterError("TTL exhausted.")

    async def execute_command_on_nodes(
        self, nodes: Iterable[Any], *args: Any, **kwargs: Any
    ) -> Any:
        command = args[0]
        res = {}

        for node in nodes:
            connection = self.connection_pool.get_connection_by_node(node)

            # copy from redis-py
            try:
                await connection.send_command(*args)
                res[node["name"]] = await self.parse_response(
                    connection, command, **kwargs
                )
            except asyncio.CancelledError:
                # do not retry when coroutine is cancelled
                connection.disconnect()
                raise
            except (ConnectionError, TimeoutError) as e:
                connection.disconnect()

                if not connection.retry_on_timeout and isinstance(e, TimeoutError):
                    raise

                await connection.send_command(*args)
                res[node["name"]] = await self.parse_response(
                    connection, command, **kwargs
                )
            finally:
                self.connection_pool.release(connection)

        return self._merge_result(command, res, **kwargs)

    def lock(
        self,
        name: StringT,
        timeout: Optional[float] = None,
        sleep: float = 0.1,
        blocking_timeout: Optional[bool] = None,
        lock_class: Optional[Type[Lock]] = None,
        thread_local: bool = True,
    ) -> Lock:
        """
        Return a new :class:`~coredis.lock.Lock` object using :paramref:`name` that mimics
        the behavior of :class:`threading.Lock`.

        See: :class:`~coredis.lock.LuaLock` (the default :paramref:`lock_class`)
         for more details.

        :raises: :exc:`~coredis.LockError`
        """

        if lock_class is None:
            if self._use_lua_lock is None:  # type: ignore
                # the first time .lock() is called, determine if we can use
                # Lua by attempting to register the necessary scripts
                try:
                    LuaLock.register_scripts(self)
                    self._use_lua_lock = True
                except ResponseError:
                    self._use_lua_lock = False
            lock_class = self._use_lua_lock and LuaLock or Lock

        return lock_class(
            self,
            name,
            timeout=timeout,
            sleep=sleep,
            blocking_timeout=blocking_timeout,
            thread_local=thread_local,
        )

    def pubsub(self, ignore_subscribe_messages: bool = False, **kwargs):
        from coredis.commands.pubsub import ClusterPubSub

        return ClusterPubSub(
            self.connection_pool,
            ignore_subscribe_messages=ignore_subscribe_messages,
            **kwargs,
        )

    async def pipeline(
        self,
        transaction: Optional[bool] = None,
        watches: Optional[Iterable[StringT]] = None,
    ) -> "coredis.commands.pipeline.ClusterPipeline":
        """
        Pipelines in cluster mode only provide a subset of the functionality
        of pipelines in standalone mode.

        Specifically:

        - Transactions are only supported if all commands in the pipeline
          only access keys which route to the same node.
        - Each command in the pipeline should only access keys on the same node
        - Transactions with :paramref:`watch` are not supported.
        """
        await self.connection_pool.initialize()

        from coredis.commands.pipeline import ClusterPipeline

        return ClusterPipeline.proxy(
            connection_pool=self.connection_pool,
            startup_nodes=self.connection_pool.nodes.startup_nodes,
            result_callbacks=self.result_callbacks,
            response_callbacks=self.response_callbacks,
            transaction=transaction,
            watches=watches,
        )

    async def transaction(self, func, *watches: StringT, **kwargs):
        """
        Convenience method for executing the callable :paramref:`func` as a
        transaction while watching all keys specified in :paramref:`watches`.
        The :paramref:`func` callable should expect a single argument which is a
        :class:`~coredis.commands.pipeline.ClusterPipeline` instance retrieved
        by calling :meth:`~coredis.RedisCluster.pipeline`

        .. warning:: Cluster transactions can only be run with commands that
           route to the same node.
        """
        value_from_callable = kwargs.pop("value_from_callable", False)
        watch_delay = kwargs.pop("watch_delay", None)
        async with await self.pipeline(True, watches=watches) as pipe:
            while True:
                try:
                    func_value = await func(pipe)
                    exec_value = await pipe.execute()
                    return func_value if value_from_callable else exec_value
                except WatchError:
                    if watch_delay is not None and watch_delay > 0:
                        await asyncio.sleep(watch_delay)
                    continue

    async def scan_iter(
        self,
        match: Optional[StringT] = None,
        count: Optional[int] = None,
        type_: Optional[StringT] = None,
    ):
        for node in self.primaries:
            cursor = "0"

            while cursor != 0:
                pieces: CommandArgList = [cursor]

                if match is not None:
                    pieces.extend([PrefixToken.MATCH, match])

                if count is not None:
                    pieces.extend([PrefixToken.COUNT, count])

                if type_ is not None:
                    pieces.extend([PrefixToken.TYPE, type_])

                response = await node.execute_command(CommandName.SCAN, *pieces)
                cursor, data = response

                for item in data:
                    yield item

    @contextlib.contextmanager
    def all(
        self,
        flag: Literal[NodeFlag.ALL, NodeFlag.PRIMARIES, NodeFlag.REPLICAS],
    ):

        if flag == NodeFlag.ALL:
            nodes = self.all_nodes
        elif flag == NodeFlag.PRIMARIES:
            nodes = self.primaries
        else:
            nodes = self.replicas
        proxy = NodeProxy(list(nodes))
        yield proxy



from __future__ import annotations

from coredis.commands import ResponseCallback
from coredis.response.utils import flat_pairs_to_dict
from coredis.typing import Any, AnyStr, Mapping, Tuple, Union


class HScanCallback(ResponseCallback):
    def transform(
        self, response: Any, **options: Any
    ) -> Tuple[int, Mapping[AnyStr, AnyStr]]:
        cursor, r = response

        return int(cursor), r and flat_pairs_to_dict(r) or {}


class HRandFieldCallback(ResponseCallback):
    def transform_3(
        self, response: Any, **options: Any
    ) -> Union[str, Tuple[AnyStr, ...], Mapping[AnyStr, AnyStr]]:
        if options.get("count"):
            if options.get("withvalues"):
                return dict(response)
            return tuple(response)
        return response

    def transform(
        self, response: Any, **options: Any
    ) -> Union[str, Tuple[AnyStr, ...], Mapping[AnyStr, AnyStr]]:
        if options.get("count"):
            if options.get("withvalues"):
                return flat_pairs_to_dict(response)
            else:
                return tuple(response)
        else:
            return response


class HGetAllCallback(ResponseCallback):
    def transform_3(self, response: Any, **options: Any) -> Mapping[AnyStr, AnyStr]:
        return response

    def transform(self, response: Any, **options: Any) -> Mapping[AnyStr, AnyStr]:
        return flat_pairs_to_dict(response) if response else {}



from __future__ import annotations

from coredis.commands import ResponseCallback
from coredis.response.callbacks import SimpleStringCallback
from coredis.response.types import LCSMatch, LCSResult
from coredis.typing import Any, AnyStr, Union
from coredis.utils import EncodingInsensitiveDict


class StringSetCallback(ResponseCallback):
    def transform(self, response: Any, **options: Any) -> Union[AnyStr, bool]:
        if options.get("get"):
            return response
        else:
            return SimpleStringCallback()(response)


class LCSCallback(ResponseCallback):
    def transform(self, response: Any, **options: Any) -> Union[AnyStr, int, LCSResult]:
        if options.get("idx") is not None:
            return LCSResult(
                tuple(
                    LCSMatch(
                        (int(k[0][0]), int(k[0][1])),
                        (int(k[1][0]), int(k[1][1])),
                        k[2] if len(k) > 2 else None,
                    )
                    for k in response[1]
                ),
                response[-1],
            )

        return response

    def transform_3(
        self, response: Any, **options: Any
    ) -> Union[AnyStr, int, LCSResult]:
        if options.get("idx") is not None:
            response_proxy = EncodingInsensitiveDict(response)
            return LCSResult(
                tuple(
                    LCSMatch(
                        (int(k[0][0]), int(k[0][1])),
                        (int(k[1][0]), int(k[1][1])),
                        k[2] if len(k) > 2 else None,
                    )
                    for k in response_proxy["matches"]
                ),
                response_proxy["len"],
            )
        return response



from __future__ import annotations

from coredis.response.callbacks import ResponseCallback
from coredis.response.utils import flat_pairs_to_dict
from coredis.typing import Any, AnyStr, Dict, List, Set, Union
from coredis.utils import EncodingInsensitiveDict


class ClientTrackingInfoCallback(ResponseCallback):
    def transform_3(
        self, response: Any, **options: Any
    ) -> Dict[AnyStr, Union[AnyStr, Set[AnyStr], List[AnyStr]]]:
        return response

    def transform(
        self, response: Any, **options: Any
    ) -> Dict[AnyStr, Union[AnyStr, Set[AnyStr], List[AnyStr]]]:
        response = EncodingInsensitiveDict(flat_pairs_to_dict(response))
        response["flags"] = set(response["flags"])
        return dict(response)



from __future__ import annotations

import weakref
from typing import AnyStr, Generic

from coredis.exceptions import FunctionError
from coredis.typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Iterable,
    KeyT,
    Optional,
    StringT,
    ValueT,
)
from coredis.utils import EncodingInsensitiveDict, nativestr

if TYPE_CHECKING:
    import coredis.client


class Library(Generic[AnyStr]):
    def __init__(
        self,
        client: coredis.client.AbstractRedis,
        name: StringT,
        code: Optional[StringT] = None,
    ):
        """
        Abstraction over a library of redis functions

        Example::

            library_code = "redis.register_function('myfunc', function(k, a) return a[1] end)"
            lib = await Library(client, "mylib", library_code)
            assert "1" == await lib["myfunc"]([], [1])
        """
        self._client: weakref.ReferenceType[coredis.client.AbstractRedis] = weakref.ref(
            client
        )
        self._name = nativestr(name)
        self._code = code
        self._functions: EncodingInsensitiveDict = EncodingInsensitiveDict()

    @property
    def client(self) -> coredis.client.AbstractRedis:
        c = self._client()
        assert c
        return c

    @property
    def functions(self) -> Dict[str, Function]:
        """
        mapping of function names to :class:`~coredis.commands.function.Function`
        instances that can be directly called.
        """
        return self._functions

    async def update(self, new_code: StringT) -> bool:
        """
        Update the code of a library with :paramref:`new_code`
        """
        if await self.client.function_load(new_code, replace=True):
            await self.__initialize()
            return True
        return False

    def __getitem__(self, function: str) -> Optional[Function]:
        return self._functions.get(function)

    async def __initialize(self):
        self._functions.clear()
        if self._code:
            await self.client.function_load(self._code)
        library = (await self.client.function_list(self._name)).get(self._name)

        if not library:
            raise FunctionError(f"No library found for {self._name}")

        for name, function in library["functions"].items():
            self._functions[name] = Function(self.client, self._name, name)

    def __await__(self):
        async def closure():
            await self.__initialize()
            return self

        return closure().__await__()


class Function:
    def __init__(
        self, client: coredis.client.AbstractRedis, library: StringT, name: StringT
    ):
        """
        Wrapper to call a redis function that has already been loaded


        Example::

            func = await Function(client, "mylib", "myfunc")
            response = await func(keys=["a"], args=[1])
        """
        self._client: weakref.ReferenceType[coredis.client.AbstractRedis] = weakref.ref(
            client
        )
        self._library = Library(client, library)
        self._name = name

    @property
    def client(self) -> coredis.client.AbstractRedis:
        c = self._client()
        assert c
        return c

    def __await__(self):
        async def closure():
            await self._library

        return closure().__await__()

    async def __call__(
        self,
        *,
        keys: Optional[Iterable[KeyT]] = None,
        args: Optional[Iterable[ValueT]] = None,
    ) -> Any:
        """
        Wrapper to call :meth:`~coredis.Redis.fcall`

        :param args:
        :param keys:
        """
        return await self.client.fcall(self._name, keys or [], args or [])



from __future__ import annotations

from coredis.response.callbacks import ResponseCallback
from coredis.response.types import LibraryDefinition
from coredis.response.utils import flat_pairs_to_dict
from coredis.typing import Any, AnyStr, Mapping, Union
from coredis.utils import EncodingInsensitiveDict


class FunctionListCallback(ResponseCallback):
    def transform(
        self, response: Any, **options: Any
    ) -> Mapping[str, LibraryDefinition]:
        libraries = [
            EncodingInsensitiveDict(flat_pairs_to_dict(library)) for library in response
        ]
        transformed = EncodingInsensitiveDict()
        for library in libraries:
            lib_name = library["library_name"]
            functions = EncodingInsensitiveDict({})
            for function in library.get("functions", []):
                function_definition = EncodingInsensitiveDict(
                    flat_pairs_to_dict(function)
                )
                functions[function_definition["name"]] = function_definition
                functions[function_definition["name"]]["flags"] = set(
                    function_definition["flags"]
                )
            library["functions"] = functions
            transformed[lib_name] = EncodingInsensitiveDict(  # type: ignore
                LibraryDefinition(
                    name=library["name"],
                    engine=library["engine"],
                    description=library["description"],
                    functions=library["functions"],
                    library_code=library["library_code"],
                )
            )
        return transformed


class FunctionStatsCallback(ResponseCallback):
    def transform(
        self, response: Any, **options: Any
    ) -> Mapping[AnyStr, Union[AnyStr, Mapping]]:
        transformed = flat_pairs_to_dict(response)
        key = b"engines" if b"engines" in transformed else "engines"
        engines = flat_pairs_to_dict(transformed.pop(key))
        for engine, stats in engines.items():
            transformed.setdefault(key, {})[engine] = flat_pairs_to_dict(stats)
        return transformed



"""
coredis.commands
----------------
"""
from __future__ import annotations

import dataclasses
import functools
import textwrap
import warnings
from abc import ABC
from types import FunctionType
from typing import TYPE_CHECKING, cast

from packaging import version

from coredis.exceptions import CommandNotSupportedError
from coredis.typing import (
    Any,
    AnyStr,
    Callable,
    Coroutine,
    Dict,
    Generic,
    NamedTuple,
    Optional,
    TypeVar,
    Union,
)

if TYPE_CHECKING:
    import coredis.client

import coredis.pool
from coredis.response.callbacks import ResponseCallback
from coredis.typing import AbstractExecutor, ParamSpec
from coredis.utils import NodeFlag

from .constants import CommandGroup, CommandName

R = TypeVar("R")
P = ParamSpec("P")


@dataclasses.dataclass
class ClusterCommandConfig:
    enabled: bool = True
    flag: Optional[NodeFlag] = None
    combine: Optional[Callable] = None

    @property
    def multi_node(self):
        return self.flag in [NodeFlag.ALL, NodeFlag.PRIMARIES]


class CommandDetails(NamedTuple):
    command: bytes
    group: Optional[CommandGroup]
    readonly: bool
    version_introduced: Optional[version.Version]
    version_deprecated: Optional[version.Version]
    arguments: Dict[str, Dict[str, str]]
    cluster: ClusterCommandConfig
    response_callback: Optional[Union[FunctionType, ResponseCallback, ResponseCallback]]


class CommandMixin(Generic[AnyStr], AbstractExecutor[AnyStr], ABC):
    connection_pool: coredis.pool.ConnectionPool


def redis_command(
    command_name: CommandName,
    group: Optional[CommandGroup] = None,
    version_introduced: Optional[str] = None,
    version_deprecated: Optional[str] = None,
    deprecation_reason: Optional[str] = None,
    arguments: Optional[Dict[str, Dict[str, str]]] = None,
    readonly: bool = False,
    response_callback: Optional[
        Union[FunctionType, ResponseCallback, ResponseCallback]
    ] = None,
    cluster: ClusterCommandConfig = ClusterCommandConfig(),
) -> Callable[
    [Callable[P, Coroutine[Any, Any, R]]], Callable[P, Coroutine[Any, Any, R]]
]:
    command_details = CommandDetails(
        command_name,
        group,
        readonly,
        version.Version(version_introduced) if version_introduced else None,
        version.Version(version_deprecated) if version_deprecated else None,
        arguments or {},
        cluster or ClusterCommandConfig(),
        response_callback,
    )

    def wrapper(
        func: Callable[P, Coroutine[Any, Any, R]]
    ) -> Callable[P, Coroutine[Any, Any, R]]:
        @functools.wraps(func)
        async def wrapped(*args: P.args, **kwargs: P.kwargs) -> R:
            _check_version(
                args[0],  # type: ignore
                command_name,
                func.__name__,
                command_details.version_introduced,
                command_details.version_deprecated,
                deprecation_reason,
            )
            return await func(*args, **kwargs)

        wrapped.__doc__ = textwrap.dedent(wrapped.__doc__ or "")
        if group:
            wrapped.__doc__ = f"""
{wrapped.__doc__}

Redis command documentation: {_redis_command_link(command_name)}
"""
        if version_introduced:
            wrapped.__doc__ += f"""
Introduced in Redis version ``{version_introduced}``
"""

        setattr(wrapped, "__coredis_command", command_details)
        return wrapped

    return wrapper


def _check_version(
    instance: Any,
    command: bytes,
    function_name: str,
    min_version: Optional[version.Version],
    deprecated_version: Optional[version.Version],
    deprecation_reason: Optional[str],
):
    if not any([min_version, deprecated_version]):
        return

    client = cast("coredis.client.RedisConnection", instance)

    if getattr(client, "verify_version", False):
        server_version = getattr(client, "server_version", None)
        if not server_version:
            return
        if min_version and server_version < min_version:
            raise CommandNotSupportedError(command, client.server_version)
        if deprecated_version and server_version >= deprecated_version:
            if deprecation_reason:
                warnings.warn(deprecation_reason.strip())
            else:
                warnings.warn(
                    f"{function_name}() is deprecated since redis version {deprecated_version}. "
                )


def _redis_command_link(command):
    return f'`{str(command)} <https://redis.io/commands/{str(command).lower().replace(" ", "-")}>`_'


__all__ = (
    "CommandGroup",
    "CommandName",
    "CommandMixin",
    "redis_command",
    "ClusterCommandConfig",
    "CommandDetails",
)



from __future__ import annotations

import hashlib

from coredis.exceptions import NoScriptError
from coredis.typing import Iterable, Optional, StringT, SupportsScript
from coredis.utils import b


class Script:
    """
    An executable Lua script object returned by :meth:`coredis.Redis.register_script`
    """

    def __init__(
        self,
        registered_client: SupportsScript,
        script: StringT,
    ):
        """
        :param script: The lua script that will be used by :meth:`execute`
        """
        self.registered_client = registered_client
        self.script = script
        self.sha = hashlib.sha1(b(script)).hexdigest()

    async def execute(
        self,
        keys: Optional[Iterable] = None,
        args: Optional[Iterable] = None,
        client: Optional[SupportsScript] = None,
    ):
        """
        Executes the script registered in :paramref:`Script.script`
        """
        from coredis.commands.pipeline import Pipeline

        if client is None:
            client = self.registered_client
        # make sure the Redis server knows about the script
        if isinstance(client, Pipeline):
            # make sure this script is good to go on pipeline
            client.scripts.add(self)

        try:
            return await client.evalsha(self.sha, keys=keys, args=args)
        except NoScriptError:
            # Maybe the client is pointed to a differnet server than the client
            # that created this instance?
            # Overwrite the sha just in case there was a discrepancy.
            self.sha = await client.script_load(self.script)
            return await client.evalsha(self.sha, keys=keys, args=args)



from __future__ import annotations

from coredis.response.callbacks import ResponseCallback
from coredis.response.utils import flat_pairs_to_dict
from coredis.typing import Any, Mapping, Tuple


class ModuleInfoCallback(ResponseCallback):
    def transform(self, response: Any, **options: Any) -> Tuple[Mapping, ...]:
        return tuple(flat_pairs_to_dict(mod) for mod in response)

    def transform_3(self, response: Any, **options: Any) -> Tuple[Mapping, ...]:
        return tuple(dict(r) for r in response)



from __future__ import annotations

from typing import Any, AnyStr, Dict, Optional, Tuple

from coredis.response.callbacks import ResponseCallback
from coredis.utils import nativestr

SENTINEL_STATE_TYPES = {
    "can-failover-its-master": int,
    "config-epoch": int,
    "down-after-milliseconds": int,
    "failover-timeout": int,
    "info-refresh": int,
    "last-hello-message": int,
    "last-ok-ping-reply": int,
    "last-ping-reply": int,
    "last-ping-sent": int,
    "master-link-down-time": int,
    "master-port": int,
    "num-other-sentinels": int,
    "num-slaves": int,
    "o-down-time": int,
    "pending-commands": int,
    "parallel-syncs": int,
    "port": int,
    "quorum": int,
    "role-reported-time": int,
    "s-down-time": int,
    "slave-priority": int,
    "slave-repl-offset": int,
    "voted-leader-epoch": int,
}


def pairs_to_dict_typed(response, type_info):
    it = iter(response)
    result = {}

    for key, value in zip(it, it):
        if key in type_info:
            try:
                value = type_info[key](value)
            except Exception:
                # if for some reason the value can't be coerced, just use
                # the string value
                pass
        result[key.replace("-", "_")] = value

    return result


def parse_sentinel_state(item) -> Dict[str, Any]:
    result = pairs_to_dict_typed([nativestr(k) for k in item], SENTINEL_STATE_TYPES)
    flags = set(result["flags"].split(","))

    for name, flag in (
        ("is_master", "master"),
        ("is_slave", "slave"),
        ("is_sdown", "s_down"),
        ("is_odown", "o_down"),
        ("is_sentinel", "sentinel"),
        ("is_disconnected", "disconnected"),
        ("is_master_down", "master_down"),
    ):
        result[name] = flag in flags

    return result


class PrimaryCallback(ResponseCallback):
    def transform(self, response: Any, **options: Any) -> Dict[str, Any]:
        return parse_sentinel_state(response)


class PrimariesCallback(ResponseCallback):
    def transform(self, response: Any, **options: Any) -> Dict[str, Dict[str, Any]]:
        result = {}

        for item in response:
            state = PrimaryCallback()(item)
            result[state["name"]] = state

        return result


class SentinelsStateCallback(ResponseCallback):
    def transform(self, response: Any, **options: Any) -> Tuple[Dict[str, Any], ...]:
        return tuple(parse_sentinel_state(map(nativestr, item)) for item in response)


class GetPrimaryCallback(ResponseCallback):
    def transform(self, response: Any, **options: Any) -> Optional[Tuple[AnyStr, int]]:
        return response and (response[0], int(response[1]))



from __future__ import annotations

from coredis.response.callbacks import (
    DictCallback,
    ResponseCallback,
    SimpleStringCallback,
)
from coredis.response.utils import flat_pairs_to_dict
from coredis.typing import Any, AnyStr, Dict, Tuple, Union


class ACLLogCallback(ResponseCallback):
    def transform(
        self, response: Any, **options: Any
    ) -> Union[bool, Tuple[Dict[AnyStr, AnyStr], ...]]:
        if options.get("reset"):
            return SimpleStringCallback()(response)
        else:
            return tuple(
                DictCallback(transform_function=flat_pairs_to_dict)(r) for r in response
            )



from __future__ import annotations

import datetime
from typing import ClassVar

from coredis.commands import ResponseCallback
from coredis.response.types import ClientInfo, RoleInfo, SlowLogInfo
from coredis.response.utils import flat_pairs_to_dict
from coredis.typing import Any, AnyStr, List, Mapping, Tuple, Union
from coredis.utils import EncodingInsensitiveDict, nativestr


class TimeCallback(ResponseCallback):
    def transform(self, response: Any, **options: Any) -> datetime.datetime:
        return datetime.datetime.fromtimestamp(int(response[0])) + datetime.timedelta(
            microseconds=int(response[1]) / 1000.0
        )


class SlowlogCallback(ResponseCallback):
    def transform(self, response: Any, **options: Any) -> Tuple[SlowLogInfo, ...]:
        return tuple(
            SlowLogInfo(
                id=item[0],
                start_time=int(item[1]),
                duration=int(item[2]),
                command=item[3],
                client_addr=item[4],
                client_name=item[5],
            )
            for item in response
        )


class ClientInfoCallback(ResponseCallback):
    INT_FIELDS: ClassVar = {
        "id",
        "fd",
        "age",
        "idle",
        "db",
        "sub",
        "psub",
        "multi",
        "qbuf-free",
        "argv-mem",
        "multi-mem",
        "obl",
        "oll",
        "omem",
        "tot-mem",
        "redir",
    }

    def transform(self, response: Any, **options: Any) -> ClientInfo:
        decoded_response = nativestr(response)
        pairs = [pair.split("=", 1) for pair in decoded_response.strip().split(" ")]

        info: ClientInfo = {}  # type: ignore
        for k, v in pairs:
            if k in ClientInfoCallback.INT_FIELDS:
                info[k] = int(v)  # type: ignore
            else:
                info[k] = v  # type: ignore
        return info


class ClientListCallback(ResponseCallback):
    def transform(self, response: Any, **options: Any) -> Tuple[ClientInfo, ...]:
        return tuple(ClientInfoCallback()(c) for c in response.splitlines())


class DebugCallback(ResponseCallback):
    INT_FIELDS: ClassVar = {"refcount", "serializedlength", "lru", "lru_seconds_idle"}

    def transform(
        self, response: Any, **options: Any
    ) -> Mapping[AnyStr, Union[AnyStr, int]]:
        # The 'type' of the object is the first item in the response, but isn't
        # prefixed with a name
        response = nativestr(response)
        response = "type:" + response
        response = dict([kv.split(":") for kv in response.split()])

        # parse some expected int values from the string response
        # note: this cmd isn't spec'd so these may not appear in all redis versions

        for field in DebugCallback.INT_FIELDS:
            if field in response:
                response[field] = int(response[field])

        return response


class InfoCallback(ResponseCallback):
    def transform(self, response: Any, **options: Any) -> Mapping[AnyStr, List[AnyStr]]:
        """Parses the result of Redis's INFO command into a Python dict"""
        info = {}
        response = nativestr(response)

        def get_value(value):
            if "," not in value or "=" not in value:
                try:
                    if "." in value:
                        return float(value)
                    else:
                        return int(value)
                except ValueError:
                    return value
            else:
                sub_dict = {}

                for item in value.split(","):
                    k, v = item.rsplit("=", 1)
                    sub_dict[k] = get_value(v)

                return sub_dict

        for line in response.splitlines():
            if line and not line.startswith("#"):
                if line.find(":") != -1:
                    key, value = line.split(":", 1)
                    info[key] = get_value(value)
                else:
                    # if the line isn't splittable, append it to the "__raw__" key
                    info.setdefault("__raw__", []).append(line)

        return info


class RoleCallback(ResponseCallback):
    def transform(self, response: Any, **options: Any) -> RoleInfo:
        role = nativestr(response[0])

        def _parse_master(response):
            offset, slaves = response[1:]
            res = {"role": role, "offset": offset, "slaves": []}

            for slave in slaves:
                host, port, offset = slave
                res["slaves"].append(
                    {"host": host, "port": int(port), "offset": int(offset)}
                )

            return res

        def _parse_slave(response):
            host, port, status, offset = response[1:]

            return dict(
                role=role,
                host=host,
                port=port,
                status=status,
                offset=offset,
            )

        def _parse_sentinel(response):
            return RoleInfo(role=role, masters=response[1:])

        parser = {
            "master": _parse_master,
            "slave": _parse_slave,
            "sentinel": _parse_sentinel,
        }[role]
        return RoleInfo(**parser(response))  # type: ignore


class LatencyHistogramCallback(ResponseCallback):
    def transform(self, response: Any, **options: Any) -> Mapping[AnyStr, Mapping]:
        histogram = flat_pairs_to_dict(response)
        for key, value in histogram.items():
            histogram[key] = EncodingInsensitiveDict(flat_pairs_to_dict(value))
            histogram[key]["histogram_usec"] = flat_pairs_to_dict(
                histogram[key]["histogram_usec"]
            )
        return histogram



from __future__ import annotations

from coredis.response.callbacks import SimpleStringCallback
from coredis.response.callbacks.sentinel import (
    GetPrimaryCallback,
    PrimariesCallback,
    PrimaryCallback,
    SentinelsStateCallback,
)
from coredis.typing import Any, AnyStr, Dict, Tuple, ValueT

from . import CommandMixin, redis_command
from .constants import CommandName


class SentinelCommands(CommandMixin[AnyStr]):
    @redis_command(
        CommandName.SENTINEL_GET_MASTER_ADDR_BY_NAME,
        response_callback=GetPrimaryCallback(),
    )
    async def sentinel_get_master_addr_by_name(self, service_name: ValueT):
        """Returns a (host, port) pair for the given :paramref:`service_name`"""

        return await self.execute_command(
            CommandName.SENTINEL_GET_MASTER_ADDR_BY_NAME, service_name
        )

    @redis_command(
        CommandName.SENTINEL_MASTER,
        response_callback=PrimaryCallback(),
    )
    async def sentinel_master(self, service_name: ValueT) -> Dict[str, Any]:
        """Returns a dictionary containing the specified masters state."""

        return await self.execute_command(CommandName.SENTINEL_MASTER, service_name)

    @redis_command(
        CommandName.SENTINEL_MASTERS,
        response_callback=PrimariesCallback(),
    )
    async def sentinel_masters(self) -> Dict[str, Dict[str, Any]]:
        """Returns a list of dictionaries containing each master's state."""

        return await self.execute_command(CommandName.SENTINEL_MASTERS)

    @redis_command(
        CommandName.SENTINEL_MONITOR,
        response_callback=SimpleStringCallback(),
    )
    async def sentinel_monitor(
        self, name: ValueT, ip: ValueT, port: int, quorum
    ) -> bool:
        """Adds a new master to Sentinel to be monitored"""

        return await self.execute_command(
            CommandName.SENTINEL_MONITOR, name, ip, port, quorum
        )

    @redis_command(
        CommandName.SENTINEL_REMOVE,
        response_callback=SimpleStringCallback(),
    )
    async def sentinel_remove(self, name: ValueT) -> bool:
        """Removes a master from Sentinel's monitoring"""

        return await self.execute_command(CommandName.SENTINEL_REMOVE, name)

    @redis_command(
        CommandName.SENTINEL_SENTINELS,
        response_callback=SentinelsStateCallback(),
    )
    async def sentinel_sentinels(
        self, service_name: ValueT
    ) -> Tuple[Dict[AnyStr, Any], ...]:
        """Returns a list of sentinels for :paramref:`service_name`"""

        return await self.execute_command(CommandName.SENTINEL_SENTINELS, service_name)

    @redis_command(
        CommandName.SENTINEL_SET,
        response_callback=SimpleStringCallback(),
    )
    async def sentinel_set(self, name: ValueT, option, value) -> bool:
        """Sets Sentinel monitoring parameters for a given master"""

        return await self.execute_command(CommandName.SENTINEL_SET, name, option, value)

    @redis_command(
        CommandName.SENTINEL_SLAVES,
        response_callback=SentinelsStateCallback(),
    )
    async def sentinel_slaves(
        self, service_name: ValueT
    ) -> Tuple[Dict[AnyStr, Any], ...]:
        """Returns a list of slaves for paramref:`service_name`"""

        return await self.execute_command(CommandName.SENTINEL_SLAVES, service_name)

    @redis_command(
        CommandName.SENTINEL_REPLICAS,
        response_callback=SentinelsStateCallback(),
    )
    async def sentinel_replicas(
        self, service_name: ValueT
    ) -> Tuple[Dict[AnyStr, Any], ...]:
        """Returns a list of replicas for :paramref:`service_name`"""

        return await self.execute_command(CommandName.SENTINEL_REPLICAS, service_name)



from __future__ import annotations

from coredis.commands import ResponseCallback
from coredis.response.types import ScoredMember, ScoredMembers
from coredis.typing import Any, AnyStr, Optional, Tuple, Union

VALID_ZADD_OPTIONS = {"NX", "XX", "CH", "INCR"}


class ZMembersOrScoredMembers(ResponseCallback):
    def transform_3(
        self, response: Any, **kwargs: Any
    ) -> Tuple[Union[AnyStr, ScoredMember], ...]:
        if kwargs.get("withscores"):
            return tuple(ScoredMember(*v) for v in response)
        else:
            return tuple(response)

    def transform(
        self, response: Any, **kwargs: Any
    ) -> Tuple[Union[AnyStr, ScoredMember], ...]:
        if not response:
            return ()
        elif kwargs.get("withscores"):
            it = iter(response)
            return tuple(ScoredMember(*v) for v in zip(it, map(float, it)))
        else:
            return tuple(response)


class ZSetScorePairCallback(ResponseCallback):
    def transform_3(
        self, response: Any, **kwargs: Any
    ) -> Optional[Union[ScoredMember, ScoredMembers]]:
        if not (kwargs.get("withscores") or kwargs.get("count")):
            return ScoredMember(*response)
        return tuple(ScoredMember(*v) for v in response)

    def transform(
        self, response: Any, **kwargs: Any
    ) -> Optional[Union[ScoredMember, ScoredMembers]]:
        if not response:
            return None
        elif not (kwargs.get("withscores") or kwargs.get("count")):
            if isinstance(response, list):
                return ScoredMember(response[0], float(response[1]))
            else:
                return response
        it = iter(response)

        return tuple(ScoredMember(*v) for v in zip(it, map(float, it)))


class ZMPopCallback(ResponseCallback):
    def transform(
        self, response: Any, **options: Any
    ) -> Optional[Tuple[AnyStr, ScoredMembers]]:
        if response:
            return (
                response[0],
                tuple(ScoredMember(v[0], float(v[1])) for v in response[1]),
            )
        return None


class ZMScoreCallback(ResponseCallback):
    def transform(self, response: Any, **options: Any) -> Tuple[Optional[float], ...]:
        return tuple(float(score) if score is not None else None for score in response)


class ZScanCallback(ResponseCallback):
    def transform(self, response: Any, **options: Any) -> Tuple[int, ScoredMembers]:
        cursor, r = response
        it = iter(r)

        return int(cursor), tuple(ScoredMember(*v) for v in zip(it, map(float, it)))


class ZRandMemberCallback(ResponseCallback):
    def transform_3(self, response: Any, **kwargs: Any) -> Union[AnyStr, ScoredMembers]:
        if not response or not kwargs.get("withscores"):
            return response
        return tuple(ScoredMember(*v) for v in response)

    def transform(self, response: Any, **kwargs: Any) -> Union[AnyStr, ScoredMembers]:
        if not response or not kwargs.get("withscores"):
            return response
        it = iter(response)

        return tuple(ScoredMember(*v) for v in zip(it, map(float, it)))


class BZPopCallback(ResponseCallback):
    def transform(
        self, response: Any, **options: Any
    ) -> Optional[Tuple[AnyStr, AnyStr, float]]:
        return response and (response[0], response[1], float(response[2]))


class ZAddCallback(ResponseCallback):
    def transform_3(
        self, response: Any, *args: Any, **kwargs: Any
    ) -> Union[int, float]:
        return response

    def transform(self, response: Any, *args: Any, **kwargs: Any) -> Union[int, float]:
        if kwargs.get("condition"):
            return float(response)
        return int(response)



from __future__ import annotations

from coredis.commands import ResponseCallback
from coredis.response.types import GeoCoordinates, GeoSearchResult
from coredis.typing import Any, List, Optional, Tuple, Union


class GeoSearchCallback(ResponseCallback):
    def transform(
        self, response: Any, **options: Any
    ) -> Union[int, Tuple[Union[str, GeoSearchResult], ...]]:
        if options.get("store") or options.get("storedist"):
            return response

        response_list = response if isinstance(response, list) else [response]

        if not (options["withdist"] or options["withcoord"] or options["withhash"]):
            return tuple(list(response_list))

        results: List[GeoSearchResult] = []

        for result in response_list:
            results.append(
                GeoSearchResult(
                    result.pop(0),
                    float(result.pop(0)) if options.get("withdist") else None,
                    result.pop(0) if options.get("withhash") else None,
                    GeoCoordinates(*map(float, result.pop(0)))
                    if options.get("withcoord")
                    else None,
                )
            )

        return tuple(results)


class GeoCoordinatessCallback(ResponseCallback):
    def transform(
        self, response: Any, **options: Any
    ) -> Tuple[Optional[GeoCoordinates], ...]:
        return tuple(
            map(
                lambda ll: (
                    GeoCoordinates(float(ll[0]), float(ll[1]))
                    if ll is not None
                    else None
                ),
                response,
            )
        )



from __future__ import annotations

from coredis.typing import (
    Iterable,
    List,
    Mapping,
    MutableMapping,
    OrderedDict,
    Tuple,
    TypeVar,
    Union,
)

T = TypeVar("T")
U = TypeVar("U")
V = TypeVar("V")
W = TypeVar("W")


def flat_pairs_to_dict(
    response: Union[Mapping[T, T], Tuple[T, ...], List[T]]
) -> MutableMapping[T, T]:
    """Creates a dict given a flat list of key/value pairs"""
    if isinstance(response, MutableMapping):
        return response
    it = iter(response)
    return dict(zip(it, it))


def flat_pairs_to_ordered_dict(response: Tuple[T, ...]) -> OrderedDict[T, T]:
    """Creates a dict given a flat list of key/value pairs"""
    it = iter(response)
    return OrderedDict(zip(it, it))


def pairs_to_dict(
    response: Union[Mapping[T, T], Tuple[Tuple[T, T], ...]]
) -> MutableMapping[T, T]:
    """Creates a dict given an array of tuples"""
    if isinstance(response, MutableMapping):
        return response
    return dict(response)


def quadruples_to_dict(
    response: Iterable[Tuple[T, U, V, W]]
) -> MutableMapping[T, Tuple[U, V, W]]:
    return {k[0]: (k[1], k[2], k[3]) for k in response}



from __future__ import annotations

from coredis.response.callbacks import ResponseCallback
from coredis.response.types import Command
from coredis.response.utils import flat_pairs_to_dict, pairs_to_dict
from coredis.typing import AbstractSet, Any, AnyStr, Mapping, MutableMapping
from coredis.utils import EncodingInsensitiveDict, nativestr


class CommandCallback(ResponseCallback):
    def transform(self, response: Any, **options: Any) -> Mapping[str, Command]:
        commands: MutableMapping[str, Command] = {}

        for command in response:
            if command:
                name = nativestr(command[0])

                if len(command) >= 6:
                    commands[name] = {
                        "name": command[0],
                        "arity": command[1],
                        "flags": command[2],
                        "first-key": command[3],
                        "last-key": command[4],
                        "step": command[5],
                        "acl-categories": None,
                        "tips": None,
                        "key-specifications": None,
                        "sub-commands": None,
                    }

                if len(command) >= 7:
                    commands[name]["acl-categories"] = command[6]

                if len(command) >= 8:
                    commands[name]["tips"] = command[7]
                    commands[name]["key-specifications"] = command[8]
                    commands[name]["sub-commands"] = command[9]

        return commands


class CommandKeyFlagCallback(ResponseCallback):
    def transform(
        self, response: Any, **options: Any
    ) -> Mapping[AnyStr, AbstractSet[AnyStr]]:
        return {k[0]: set(k[1]) for k in response}

    def transform_3(
        self, response: Any, **options: Any
    ) -> Mapping[AnyStr, AbstractSet[AnyStr]]:
        return pairs_to_dict(response)


class CommandDocCallback(ResponseCallback):
    def transform(self, response: Any, **options: Any) -> Mapping[AnyStr, Mapping]:
        cmd_mapping = flat_pairs_to_dict(response)
        for cmd, doc in cmd_mapping.items():
            cmd_mapping[cmd] = EncodingInsensitiveDict(flat_pairs_to_dict(doc))
            cmd_mapping[cmd]["arguments"] = [
                flat_pairs_to_dict(arg) for arg in cmd_mapping[cmd]["arguments"]
            ]
        return cmd_mapping

    def transform_3(self, response: Any, **options: Any) -> Mapping[AnyStr, Mapping]:
        return dict(response)



from __future__ import annotations

from coredis.commands import ResponseCallback
from coredis.typing import Any, AnyStr, Dict


class NumSubCallback(ResponseCallback):
    def transform(self, response: Any, **options: Any) -> Dict[AnyStr, int]:
        return dict(zip(response[0::2], response[1::2]))



from __future__ import annotations

import os
import warnings
from abc import ABC, abstractmethod
from numbers import Number
from typing import (
    TYPE_CHECKING,
    AbstractSet,
    Any,
    AnyStr,
    AsyncGenerator,
    Awaitable,
    Callable,
    ClassVar,
    Coroutine,
    Dict,
    Generic,
    Iterable,
    Iterator,
    List,
    Literal,
    Mapping,
    MutableMapping,
    NamedTuple,
    Optional,
    Protocol,
    Sequence,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
)

from typing_extensions import (
    Deque,
    OrderedDict,
    ParamSpec,
    TypeAlias,
    TypedDict,
    runtime_checkable,
)

RUNTIME_TYPECHECKS = False

if os.environ.get("COREDIS_RUNTIME_CHECKS", "").lower() in ["1", "true", "t"]:
    try:
        import beartype

        if not TYPE_CHECKING:
            from beartype.typing import (  # noqa: F811
                Deque,
                Dict,
                Iterable,
                Iterator,
                List,
                Mapping,
                OrderedDict,
                Sequence,
                Set,
                Tuple,
                TypedDict,
            )

        RUNTIME_TYPECHECKS = True
    except ImportError:  # noqa
        warnings.warn("Runtime checks were requested but could not import beartype")

CommandArgList = List[Union[str, bytes, float, Number]]


P = ParamSpec("P")
R = TypeVar("R")

KeyT: TypeAlias = Union[str, bytes]
ValueT: TypeAlias = Union[str, bytes, int, float]
StringT: TypeAlias = KeyT

# TODO: mypy can't handle recursive types
ResponseType = Optional[
    Union[
        StringT,
        int,
        float,
        bool,
        AbstractSet,
        List,
        Tuple,
        Mapping,
        # AbstractSet["ResponseType"],
        # List["ResponseType"],
        # Mapping["ResponseType", "ResponseType"],
        Exception,
    ]
]


def add_runtime_checks(func: Callable[P, R]) -> Callable[P, R]:
    if RUNTIME_TYPECHECKS:
        return beartype.beartype(func)

    return func


@runtime_checkable
class SupportsWatch(Protocol):  # noqa
    async def __aenter__(self) -> "SupportsWatch":
        ...

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        ...

    async def watch(self, *keys: KeyT) -> bool:
        ...

    async def execute(self, raise_on_error=True) -> Any:
        ...


@runtime_checkable
class SupportsScript(Protocol):  # noqa
    async def evalsha(
        self,
        sha1: StringT,
        keys: Optional[Iterable[KeyT]] = None,
        args: Optional[Iterable[ValueT]] = None,
    ) -> Any:
        ...

    async def evalsha_ro(
        self,
        sha1: StringT,
        keys: Optional[Iterable[KeyT]] = None,
        args: Optional[Iterable[ValueT]] = None,
    ) -> Any:
        ...

    async def script_load(self, script: StringT) -> AnyStr:
        ...


@runtime_checkable
class SupportsPipeline(Protocol):  # noqa
    async def pipeline(
        self,
        transaction: Optional[bool] = True,
        watches: Optional[Iterable[StringT]] = None,
    ) -> SupportsWatch:
        ...


class AbstractExecutor(ABC, Generic[AnyStr]):
    @abstractmethod
    async def execute_command(self, command: bytes, *args: Any, **options: Any) -> Any:
        pass


__all__ = [
    "AbstractExecutor",
    "AbstractSet",
    "Any",
    "AnyStr",
    "AsyncGenerator",
    "Awaitable",
    "Callable",
    "ClassVar",
    "CommandArgList",
    "Coroutine",
    "Deque",
    "Dict",
    "Generic",
    "KeyT",
    "Iterable",
    "Iterator",
    "List",
    "Literal",
    "Mapping",
    "MutableMapping",
    "NamedTuple",
    "OrderedDict",
    "Optional",
    "ParamSpec",
    "Protocol",
    "ResponseType",
    "Sequence",
    "Set",
    "SupportsWatch",
    "SupportsScript",
    "SupportsPipeline",
    "StringT",
    "Tuple",
    "Type",
    "TypedDict",
    "TypeVar",
    "Union",
    "ValueT",
    "TYPE_CHECKING",
    "RUNTIME_TYPECHECKS",
]
